import datetime
import time
from typing import Iterable, Optional, TYPE_CHECKING, Union

from bxcommon import constants
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.feed.feed_source import FeedSource
from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.messages.bloxroute.block_holding_message import BlockHoldingMessage
from bxcommon.messages.bloxroute.get_txs_message import GetTxsMessage
from bxcommon.messages.eth.validation.abstract_block_validator import AbstractBlockValidator, \
    BlockValidationResult
from bxcommon.utils import convert, crypto, block_content_debug_utils
from bxcommon.utils.blockchain_utils.eth import eth_common_utils
from bxcommon.utils.expiring_dict import ExpiringDict
from bxcommon.utils.limited_size_set import LimitedSizeSet
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.utils.stats.stat_block_type import StatBlockType
from bxcommon.utils.stats.transaction_stat_event_type import TransactionStatEventType
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxgateway import gateway_constants
from bxgateway import log_messages
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxgateway.messages.gateway.block_received_message import BlockReceivedMessage
from bxgateway.services.block_recovery_service import BlockRecoveryInfo, RecoveredTxsSource
from bxgateway.utils.errors.message_conversion_error import MessageConversionError
from bxgateway.utils.stats.gateway_bdn_performance_stats_service import gateway_bdn_performance_stats_service
from bxutils import logging

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(__name__)


class BlockHold:
    """
    Data class for holds on block messages.

    Attributes
    ----------
    hold_message_time: time the hold message was received. Currently unused, for use later to determine if Gateway is
                       sending malicious hold messages
    holding_connection: connection that sent the original hold message
    block_message: message being delayed due to an existing hold
    alarm: reference to alarm object for hold timeout
    connection: connection from which the message is being delayed for
    """

    def __init__(self, hold_message_time, holding_connection):
        self.hold_message_time = hold_message_time
        self.holding_connection = holding_connection

        self.block_message = None
        self.alarm = None
        self.connection = None


class BlockProcessingService:
    """
    Service class that process blocks.
    Blocks received from blockchain node are held if gateway receives a `blockhold` message from another gateway to
    prevent duplicate message sending.
    """

    def __init__(self, node):
        self._node: AbstractGatewayNode = node
        self._holds = ExpiringDict(
            node.alarm_queue,
            node.opts.blockchain_block_hold_timeout_s,
            f"block_processing_holds"
        )

        self._block_validator: Optional[AbstractBlockValidator] = None
        self._last_confirmed_block_number: Optional[int] = None
        self._last_confirmed_block_difficulty: Optional[int] = None
        self._blocks_failed_validation_history: LimitedSizeSet[Sha256Hash] = LimitedSizeSet(
            constants.BLOCKS_FAILED_VALIDATION_HISTORY_SIZE)

    def place_hold(self, block_hash, connection) -> None:
        """
        Places hold on block hash and propagates message.
        :param block_hash: ObjectHash
        :param connection:
        """
        block_stats.add_block_event_by_block_hash(
            block_hash, BlockStatEventType.BLOCK_HOLD_REQUESTED,
            network_num=connection.network_num,
            peers=[connection],
        )

        if block_hash in self._node.blocks_seen.contents:
            return

        if block_hash not in self._holds.contents:
            self._holds.add(block_hash, BlockHold(time.time(), connection))
            conns = self._node.broadcast(
                BlockHoldingMessage(block_hash, self._node.network_num),
                broadcasting_conn=connection,
                connection_types=(ConnectionType.RELAY_BLOCK, ConnectionType.GATEWAY)
            )
            if len(conns) > 0:
                block_stats.add_block_event_by_block_hash(
                    block_hash,
                    BlockStatEventType.BLOCK_HOLD_SENT_BY_GATEWAY_TO_PEERS,
                    network_num=self._node.network_num,
                    peers=conns,
                )

    def queue_block_for_processing(self, block_message, connection) -> None:
        """
        Queues up block for processing on timeout if hold message received.
        If no hold exists, compress and broadcast block immediately.
        :param block_message: block message to process
        :param connection: receiving connection (AbstractBlockchainConnection)
        """

        block_hash = block_message.block_hash()
        connection.log_info(
            "Processing block {} from local blockchain node.",
            block_hash
        )

        valid_block = self._validate_block_header_in_block_message(block_message)
        if not valid_block.is_valid:
            reason = valid_block.reason
            assert reason is not None
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE_FAILED_VALIDATION,
                connection.network_num,
                more_info=reason
            )
            return

        if block_hash in self._holds.contents:
            hold: BlockHold = self._holds.contents[block_hash]
            block_stats.add_block_event_by_block_hash(
                block_hash, BlockStatEventType.BLOCK_HOLD_HELD_BLOCK,
                network_num=connection.network_num,
                peers=[hold.holding_connection],
            )
            if hold.alarm is None:
                hold.alarm = self._node.alarm_queue.register_alarm(
                    self._node.opts.blockchain_block_hold_timeout_s, self._holding_timeout, block_hash, hold
                )
                hold.block_message = block_message
                hold.connection = connection
        else:
            if self._node.opts.encrypt_blocks:
                # Broadcast holding message if gateway wants to encrypt blocks
                conns = self._node.broadcast(
                    BlockHoldingMessage(block_hash, self._node.network_num),
                    broadcasting_conn=connection,
                    prepend_to_queue=True,
                    connection_types=(ConnectionType.RELAY_BLOCK, ConnectionType.GATEWAY)
                )
                if len(conns) > 0:
                    block_stats.add_block_event_by_block_hash(
                        block_hash,
                        BlockStatEventType.BLOCK_HOLD_SENT_BY_GATEWAY_TO_PEERS,
                        network_num=self._node.network_num,
                        peers=conns
                    )
            self._process_and_broadcast_block(block_message, connection)

    def cancel_hold_timeout(self, block_hash, connection) -> None:
        """
        Lifts hold on block hash and cancels timeout.
        :param block_hash: ObjectHash
        :param connection: connection cancelling hold
        """
        if block_hash in self._holds.contents:
            block_stats.add_block_event_by_block_hash(block_hash, BlockStatEventType.BLOCK_HOLD_LIFTED,
                                                      network_num=connection.network_num,
                                                      peers=[connection],
                                                      )

            hold = self._holds.contents[block_hash]
            if hold.alarm is not None:
                self._node.alarm_queue.unregister_alarm(hold.alarm)
            del self._holds.contents[block_hash]

    def process_block_broadcast(self, msg, connection: AbstractRelayConnection) -> None:
        """
        Handle broadcast message receive from bloXroute.
        This is typically an encrypted block.
        """

        # TODO handle the situation where txs that received from relays while syncing are in the blocks that were
        #  ignored while syncing, so these txs won't be cleaned for 3 days
        if not self._node.should_process_block_hash(msg.block_hash()):
            return

        block_stats.add_block_event(
            msg,
            BlockStatEventType.ENC_BLOCK_RECEIVED_BY_GATEWAY_FROM_NETWORK,
            network_num=connection.network_num,
            more_info=stats_format.connection(connection)
        )

        block_hash = msg.block_hash()
        is_encrypted = msg.is_encrypted()
        self._node.track_block_from_bdn_handling_started(block_hash, connection.peer_desc)

        if not is_encrypted:
            block = msg.blob()
            self._handle_decrypted_block(block, connection)
            return

        cipherblob = msg.blob()
        expected_hash = Sha256Hash(crypto.double_sha256(cipherblob))
        if block_hash != expected_hash:
            connection.log_warning(log_messages.BLOCK_WITH_INCONSISTENT_HASHES,
                                   expected_hash, block_hash)
            return

        if self._node.in_progress_blocks.has_encryption_key_for_hash(block_hash):
            connection.log_trace("Already had key for received block. Sending block to node.")
            decrypt_start_timestamp = time.time()
            decrypt_start_datetime = datetime.datetime.utcnow()
            block = self._node.in_progress_blocks.decrypt_ciphertext(block_hash, cipherblob)

            if block is not None:
                block_stats.add_block_event(
                    msg,
                    BlockStatEventType.ENC_BLOCK_DECRYPTED_SUCCESS,
                    start_date_time=decrypt_start_datetime,
                    end_date_time=datetime.datetime.utcnow(),
                    network_num=connection.network_num,
                    more_info=stats_format.timespan(decrypt_start_timestamp, time.time())
                )
                self._handle_decrypted_block(
                    block,
                    connection,
                    encrypted_block_hash_hex=convert.bytes_to_hex(block_hash.binary)
                )
            else:
                block_stats.add_block_event(msg,
                                            BlockStatEventType.ENC_BLOCK_DECRYPTION_ERROR,
                                            network_num=connection.network_num)
        else:
            connection.log_trace("Received encrypted block. Storing.")
            self._node.in_progress_blocks.add_ciphertext(block_hash, cipherblob)
            block_received_message = BlockReceivedMessage(block_hash)
            conns = self._node.broadcast(
                block_received_message,
                connection,
                connection_types=(ConnectionType.GATEWAY,)
            )
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.ENC_BLOCK_SENT_BLOCK_RECEIPT,
                network_num=connection.network_num,
                peers=conns,
            )

    def process_block_key(self, msg, connection: AbstractRelayConnection) -> None:
        """
        Handles key message receive from bloXroute.
        Looks for the encrypted block and decrypts; otherwise stores for later.
        """
        key = msg.key()
        block_hash = msg.block_hash()

        if not self._node.should_process_block_hash(block_hash):
            return

        block_stats.add_block_event_by_block_hash(block_hash,
                                                  BlockStatEventType.ENC_BLOCK_KEY_RECEIVED_BY_GATEWAY_FROM_NETWORK,
                                                  network_num=connection.network_num,
                                                  peers=[connection])

        if self._node.in_progress_blocks.has_encryption_key_for_hash(block_hash):
            return

        if self._node.in_progress_blocks.has_ciphertext_for_hash(block_hash):
            connection.log_trace("Cipher text found. Decrypting and sending to node.")
            decrypt_start_timestamp = time.time()
            decrypt_start_datetime = datetime.datetime.utcnow()
            block = self._node.in_progress_blocks.decrypt_and_get_payload(block_hash, key)

            if block is not None:
                block_stats.add_block_event_by_block_hash(
                    block_hash,
                    BlockStatEventType.ENC_BLOCK_DECRYPTED_SUCCESS,
                    start_date_time=decrypt_start_datetime,
                    end_date_time=datetime.datetime.utcnow(),
                    network_num=connection.network_num,
                    more_info=stats_format.timespan(
                        decrypt_start_timestamp,
                        time.time())
                )
                self._handle_decrypted_block(
                    block, connection,
                    encrypted_block_hash_hex=convert.bytes_to_hex(block_hash.binary)
                )
            else:
                block_stats.add_block_event_by_block_hash(
                    block_hash,
                    BlockStatEventType.ENC_BLOCK_DECRYPTION_ERROR,
                    network_num=connection.network_num
                )
        else:
            connection.log_trace("No cipher text found on key message. Storing.")
            self._node.in_progress_blocks.add_key(block_hash, key)

        conns = self._node.broadcast(msg, connection, connection_types=(ConnectionType.GATEWAY,))
        if len(conns) > 0:
            block_stats.add_block_event_by_block_hash(block_hash,
                                                      BlockStatEventType.ENC_BLOCK_KEY_SENT_BY_GATEWAY_TO_PEERS,
                                                      network_num=self._node.network_num,
                                                      peers=conns
                                                      )

    def retry_broadcast_recovered_blocks(self, connection) -> None:
        if self._node.block_recovery_service.recovered_blocks and self._node.opts.has_fully_updated_tx_service:
            for msg, recovery_source in self._node.block_recovery_service.recovered_blocks:
                self._handle_decrypted_block(msg, connection, recovered=True, recovered_txs_source=recovery_source)

            self._node.block_recovery_service.clean_up_recovered_blocks()

    def reset_last_confirmed_block_parameters(self):
        self._last_confirmed_block_number = None
        self._last_confirmed_block_difficulty = None

    def set_last_confirmed_block_parameters(
        self,
        last_confirmed_block_number: int,
        last_confirmed_block_difficulty: int
    ) -> None:

        if not self._node.peer_relays:
            logger.debug("Skip updating last confirmed block parameters because there is no connection to block relay")
            return

        if self._last_confirmed_block_number is not None:

            old_last_confirmed_block_number = self._last_confirmed_block_number
            assert old_last_confirmed_block_number is not None

            if last_confirmed_block_number < old_last_confirmed_block_number:
                logger.trace("New last confirmed block number {} is smaller than current {}. Skipping operation.",
                             last_confirmed_block_number, old_last_confirmed_block_number)
                return

        self._last_confirmed_block_number = last_confirmed_block_number
        self._last_confirmed_block_difficulty = last_confirmed_block_difficulty

        logger.trace("Updated last confirmed block number to {} and difficulty to {}.",
                     self._last_confirmed_block_number, self._last_confirmed_block_difficulty)

    def _compute_hold_timeout(self, _block_message) -> int:
        """
        Computes timeout after receiving block message before sending the block anyway if not received from network.
        TODO: implement algorithm for computing hold timeout
        :param block_message: block message to hold
        :return: time in seconds to wait
        """
        return self._node.opts.blockchain_block_hold_timeout_s

    def _holding_timeout(self, block_hash, hold):
        block_stats.add_block_event_by_block_hash(block_hash, BlockStatEventType.BLOCK_HOLD_TIMED_OUT,
                                                  network_num=hold.connection.network_num,
                                                  peers=[hold.connection]
                                                  )
        self._process_and_broadcast_block(hold.block_message, hold.connection)

    def _process_and_broadcast_block(self, block_message, connection: AbstractGatewayBlockchainConnection) -> None:
        """
        Compresses and propagates block message if enabled, else return.
        :param block_message: block message to propagate
        :param connection: receiving connection (AbstractBlockchainConnection)
        """
        block_hash = block_message.block_hash()
        message_converter = self._node.message_converter
        assert message_converter is not None
        try:
            bx_block, block_info = message_converter.block_to_bx_block(
                block_message,
                self._node.get_tx_service(),
                self._node.opts.enable_block_compression,
                self._node.network.min_tx_age_seconds
            )
        except MessageConversionError as e:
            block_stats.add_block_event_by_block_hash(
                e.msg_hash,
                BlockStatEventType.BLOCK_CONVERSION_FAILED,
                network_num=connection.network_num,
                conversion_type=e.conversion_type.value
            )
            connection.log_error(log_messages.BLOCK_COMPRESSION_FAIL, e.msg_hash, e)
            return

        if block_info.ignored_short_ids:
            assert block_info.ignored_short_ids is not None
            logger.debug(
                "Ignoring {} new SIDs for {}: {}",
                len(block_info.ignored_short_ids), block_info.block_hash, block_info.ignored_short_ids
            )

        compression_rate = block_info.compression_rate
        assert compression_rate is not None
        block_stats.add_block_event_by_block_hash(
            block_hash,
            BlockStatEventType.BLOCK_COMPRESSED,
            start_date_time=block_info.start_datetime,
            end_date_time=block_info.end_datetime,
            network_num=connection.network_num,
            prev_block_hash=block_info.prev_block_hash,
            original_size=block_info.original_size,
            txs_count=block_info.txn_count,
            blockchain_network=self._node.opts.blockchain_network,
            blockchain_protocol=self._node.opts.blockchain_protocol,
            matching_block_hash=block_info.compressed_block_hash,
            matching_block_type=StatBlockType.COMPRESSED.value,
            more_info="Compression: {}->{} bytes, {}, {}; Tx count: {}".format(
                block_info.original_size,
                block_info.compressed_size,
                stats_format.percentage(compression_rate),
                stats_format.duration(block_info.duration_ms),
                block_info.txn_count
            )
        )
        if self._node.opts.dump_short_id_mapping_compression:
            mapping = {}
            for short_id in block_info.short_ids:
                tx_hash = self._node.get_tx_service().get_transaction(short_id).hash
                assert tx_hash is not None
                mapping[short_id] = convert.bytes_to_hex(tx_hash.binary)
            with open(f"{self._node.opts.dump_short_id_mapping_compression_path}/"
                      f"{convert.bytes_to_hex(block_hash.binary)}", "w") as f:
                f.write(str(mapping))

        self._process_and_broadcast_compressed_block(bx_block, connection, block_info, block_hash)

        self._node.log_blocks_network_content(self._node.network_num, block_message)

    def _process_and_broadcast_compressed_block(
        self,
        bx_block,
        connection: AbstractGatewayBlockchainConnection,
        block_info,
        block_hash: Sha256Hash
    ) -> None:
        """
        Process a compressed block.
        :param bx_block: compress block message to process
        :param connection: receiving connection (AbstractBlockchainConnection)
        :param block_info: original block info
        :param block_hash: block hash
        """
        connection.node.neutrality_service.propagate_block_to_network(bx_block, connection, block_info)
        self._node.get_tx_service().track_seen_short_ids_delayed(block_hash, block_info.short_ids)

    def _handle_decrypted_block(
        self,
        bx_block: memoryview,
        connection: AbstractRelayConnection,
        encrypted_block_hash_hex: Optional[str] = None,
        recovered: bool = False,
        recovered_txs_source: Optional[RecoveredTxsSource] = None
    ) -> None:
        transaction_service = self._node.get_tx_service()
        message_converter = self._node.message_converter
        assert message_converter is not None

        valid_block = self._validate_compressed_block_header(bx_block)
        if not valid_block.is_valid:
            reason = valid_block.reason
            assert reason is not None
            block_stats.add_block_event_by_block_hash(
                valid_block.block_hash,
                BlockStatEventType.BLOCK_DECOMPRESSED_FAILED_VALIDATION,
                connection.network_num,
                more_info=reason
            )
            return

        # TODO: determine if a real block or test block. Discard if test block.
        if self._node.remote_node_conn or self._node.has_active_blockchain_peer():
            try:
                (
                    block_message, block_info, unknown_sids, unknown_hashes
                ) = message_converter.bx_block_to_block(bx_block, transaction_service)
                block_content_debug_utils.log_compressed_block_debug_info(transaction_service, bx_block)
            except MessageConversionError as e:
                block_stats.add_block_event_by_block_hash(
                    e.msg_hash,
                    BlockStatEventType.BLOCK_CONVERSION_FAILED,
                    network_num=connection.network_num,
                    conversion_type=e.conversion_type.value
                )
                transaction_service.on_block_cleaned_up(e.msg_hash)
                connection.log_warning(log_messages.FAILED_TO_DECOMPRESS_BLOCK, e.msg_hash, e)
                return
        else:
            connection.log_warning(log_messages.LACK_BLOCKCHAIN_CONNECTION)
            return

        block_hash = block_info.block_hash
        all_sids = block_info.short_ids

        if encrypted_block_hash_hex is not None:
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_TO_ENC_BLOCK_MATCH,
                matching_block_hash=encrypted_block_hash_hex,
                matching_block_type=StatBlockType.ENCRYPTED.value,
                network_num=connection.network_num
            )

        self.cancel_hold_timeout(block_hash, connection)

        if recovered:
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_RECOVERY_COMPLETED,
                network_num=connection.network_num,
                more_info=str(recovered_txs_source)
            )

        if block_hash in self._node.blocks_seen.contents:
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_DECOMPRESSED_IGNORE_SEEN,
                start_date_time=block_info.start_datetime,
                end_date_time=block_info.end_datetime,
                network_num=connection.network_num,
                prev_block_hash=block_info.prev_block_hash,
                original_size=block_info.original_size,
                compressed_size=block_info.compressed_size,
                txs_count=block_info.txn_count,
                blockchain_network=self._node.opts.blockchain_network,
                blockchain_protocol=self._node.opts.blockchain_protocol,
                matching_block_hash=block_info.compressed_block_hash,
                matching_block_type=StatBlockType.COMPRESSED.value,
                more_info=stats_format.duration(block_info.duration_ms)
            )
            self._node.track_block_from_bdn_handling_ended(block_hash)
            transaction_service.track_seen_short_ids(block_hash, all_sids)
            connection.log_info(
                "Discarding duplicate block {} from the BDN.",
                block_hash
            )
            if block_message is not None:
                self._node.on_block_received_from_bdn(block_hash, block_message)
                if self._node.block_queuing_service_manager.get_block_data(block_hash) is None:
                    self._node.block_queuing_service_manager.store_block_data(block_hash, block_message)
            return

        if not recovered:
            connection.log_info("Received block {} from the BDN.", block_hash)
        else:
            connection.log_info("Successfully recovered block {}.", block_hash)

        if block_message is not None:
            compression_rate = block_info.compression_rate
            assert compression_rate is not None
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_DECOMPRESSED_SUCCESS,
                start_date_time=block_info.start_datetime,
                end_date_time=block_info.end_datetime,
                network_num=connection.network_num,
                prev_block_hash=block_info.prev_block_hash,
                original_size=block_info.original_size,
                compressed_size=block_info.compressed_size,
                txs_count=block_info.txn_count,
                blockchain_network=self._node.opts.blockchain_network,
                blockchain_protocol=self._node.opts.blockchain_protocol,
                matching_block_hash=block_info.compressed_block_hash,
                matching_block_type=StatBlockType.COMPRESSED.value,
                more_info="Compression rate {}, Decompression time {}, "
                          "Queued behind {} blocks".format(
                    stats_format.percentage(compression_rate),
                    stats_format.duration(block_info.duration_ms),
                    self._node.block_queuing_service_manager.get_length_of_each_queuing_service_stats_format()
                )
            )

            self._on_block_decompressed(block_message)
            if recovered or self._node.block_queuing_service_manager.is_in_any_queuing_service(block_hash):
                self._node.block_queuing_service_manager.update_recovered_block(block_hash, block_message)
            else:
                self._node.block_queuing_service_manager.push(block_hash, block_message)

            gateway_bdn_performance_stats_service.log_block_from_bdn()

            self._node.on_block_received_from_bdn(block_hash, block_message)
            transaction_service.track_seen_short_ids(block_hash, all_sids)

            self._node.publish_block(None, block_hash, block_message, FeedSource.BDN_SOCKET)
            self._node.log_blocks_network_content(self._node.network_num, block_message)
        else:
            if self._node.block_queuing_service_manager.is_in_any_queuing_service(block_hash) and not recovered:
                connection.log_trace("Handling already queued block again. Ignoring.")
                return

            self._node.block_recovery_service.add_block(bx_block, block_hash, unknown_sids, unknown_hashes)
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_DECOMPRESSED_WITH_UNKNOWN_TXS,
                start_date_time=block_info.start_datetime,
                end_date_time=block_info.end_datetime,
                network_num=connection.network_num,
                prev_block_hash=block_info.prev_block_hash,
                original_size=block_info.original_size,
                compressed_size=block_info.compressed_size,
                txs_count=block_info.txn_count,
                blockchain_network=self._node.opts.blockchain_network,
                blockchain_protocol=self._node.opts.blockchain_protocol,
                matching_block_hash=block_info.compressed_block_hash,
                matching_block_type=StatBlockType.COMPRESSED.value,
                more_info="{} sids, {} hashes, [{},...]".format(len(unknown_sids), len(unknown_hashes),
                                                                unknown_sids[:5])
            )

            connection.log_info("Block {} requires short id recovery. Querying BDN...", block_hash)

            self.start_transaction_recovery(unknown_sids, unknown_hashes, block_hash, connection)
            if recovered:
                # should never happen –– this should not be called on blocks that have not recovered
                connection.log_error(log_messages.BLOCK_DECOMPRESSION_FAILURE,
                                     block_hash)
            else:
                self._node.block_queuing_service_manager.push(
                    block_hash, waiting_for_recovery=True
                )

    def start_transaction_recovery(
        self,
        unknown_sids: Iterable[int],
        unknown_hashes: Iterable[Sha256Hash],
        block_hash: Sha256Hash,
        connection: Optional[AbstractRelayConnection] = None
    ) -> None:
        all_unknown_sids = []
        all_unknown_sids.extend(unknown_sids)
        tx_service = self._node.get_tx_service()

        # retrieving sids of txs with unknown contents
        for tx_hash in unknown_hashes:
            transaction_key = tx_service.get_transaction_key(tx_hash)
            tx_sid = tx_service.get_short_id_by_key(transaction_key)
            all_unknown_sids.append(tx_sid)

        if not self._node.opts.request_recovery:
            if connection is not None:
                network_num = connection.network_num
            else:
                network_num = self._node.network_num
            # log recovery started to match with recovery completing
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_RECOVERY_STARTED,
                network_num=network_num,
                txs_count=len(all_unknown_sids),
                more_info="recovery from relay is disabled",
            )
            return

        get_txs_message = GetTxsMessage(short_ids=all_unknown_sids)
        self._node.broadcast(get_txs_message, connection_types=(ConnectionType.RELAY_TRANSACTION,))

        if connection is not None:
            tx_stats.add_txs_by_short_ids_event(
                all_unknown_sids,
                TransactionStatEventType.TX_UNKNOWN_SHORT_IDS_REQUESTED_BY_GATEWAY_FROM_RELAY,
                network_num=self._node.network_num,
                peers=[connection],
                block_hash=convert.bytes_to_hex(block_hash.binary)
            )
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_RECOVERY_STARTED,
                network_num=connection.network_num,
                txs_count=len(all_unknown_sids),
                request_hash=convert.bytes_to_hex(
                    crypto.double_sha256(get_txs_message.rawbytes())
                )
            )
        else:
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_RECOVERY_REPEATED,
                network_num=self._node.network_num,
                txs_count=len(all_unknown_sids),
                request_hash=convert.bytes_to_hex(
                    crypto.double_sha256(get_txs_message.rawbytes())
                )
            )

    def schedule_recovery_retry(self, block_awaiting_recovery: BlockRecoveryInfo) -> None:
        """
        Schedules a block recovery attempt. Repeated block recovery attempts result in longer timeouts,
        following `gateway_constants.BLOCK_RECOVERY_INTERVAL_S`'s pattern, until giving up.
        :param block_awaiting_recovery: info about recovering block
        :return:
        """
        block_hash = block_awaiting_recovery.block_hash
        recovery_attempts = self._node.block_recovery_service.recovery_attempts_by_block[block_hash]
        recovery_timed_out = time.time() - block_awaiting_recovery.recovery_start_time >= \
                             self._node.opts.blockchain_block_recovery_timeout_s
        if recovery_attempts >= gateway_constants.BLOCK_RECOVERY_MAX_RETRY_ATTEMPTS or recovery_timed_out:
            logger.error(log_messages.SHORT_ID_RECOVERY_FAIL, block_hash)
            self._node.block_recovery_service.cancel_recovery_for_block(block_hash)
            self._node.block_queuing_service_manager.remove(block_hash)
        else:
            delay = gateway_constants.BLOCK_RECOVERY_RECOVERY_INTERVAL_S[recovery_attempts]
            self._node.alarm_queue.register_approx_alarm(delay, delay / 2, self._trigger_recovery_retry,
                                                         block_awaiting_recovery)

    def _trigger_recovery_retry(self, block_awaiting_recovery: BlockRecoveryInfo) -> None:
        block_hash = block_awaiting_recovery.block_hash
        if self._node.block_recovery_service.awaiting_recovery(block_hash):
            self._node.block_recovery_service.recovery_attempts_by_block[block_hash] += 1
            self.start_transaction_recovery(
                block_awaiting_recovery.unknown_short_ids,
                block_awaiting_recovery.unknown_transaction_hashes,
                block_hash
            )

    def _on_block_decompressed(self, block_msg) -> None:
        pass

    def _validate_block_header_in_block_message(
        self, block_message: AbstractBlockMessage
    ) -> BlockValidationResult:
        block_header_bytes = self._get_block_header_bytes_from_block_message(block_message)
        validation_result = self._validate_block_header(block_header_bytes)

        if not validation_result.is_valid and validation_result.block_hash:
            block_hash = validation_result.block_hash
            assert block_hash is not None
            if block_hash in self._blocks_failed_validation_history:
                block_number = eth_common_utils.block_header_number(block_header_bytes)
                block_difficulty = eth_common_utils.block_header_difficulty(block_header_bytes)
                self.set_last_confirmed_block_parameters(block_number, block_difficulty)

        return validation_result

    def _validate_compressed_block_header(
        self, compressed_block_bytes: Union[bytearray, memoryview]
    ) -> BlockValidationResult:
        block_header_bytes = self._get_compressed_block_header_bytes(compressed_block_bytes)
        validation_result = self._validate_block_header(block_header_bytes)
        if not validation_result.is_valid and validation_result.block_hash:
            block_hash = validation_result.block_hash
            assert block_hash is not None
            self._blocks_failed_validation_history.add(block_hash)
        return validation_result

    def _validate_block_header(self, block_header_bytes: Union[bytearray, memoryview]) -> BlockValidationResult:
        if self._block_validator and self._last_confirmed_block_number and self._last_confirmed_block_difficulty:
            block_validator = self._block_validator
            assert block_validator is not None
            return block_validator.validate_block_header(
                block_header_bytes,
                self._last_confirmed_block_number,
                self._last_confirmed_block_difficulty
            )
        logger.debug(
            "Skipping block validation. Block validator - {}, last confirmed block - {}, last confirmed block difficulty - {}",
            self._block_validator, self._last_confirmed_block_number, self._last_confirmed_block_difficulty)
        return BlockValidationResult(True, None, None)

    def _get_compressed_block_header_bytes(
        self, compressed_block_bytes: Union[bytearray, memoryview]
    ) -> Union[bytearray, memoryview]:
        pass

    def _get_block_header_bytes_from_block_message(
        self, block_message: AbstractBlockMessage
    ) -> Union[bytearray, memoryview]:
        pass



from bxgateway.testing import gateway_helpers
from bxcommon.test_utils.abstract_test_case import AbstractTestCase
from bxcommon import constants
from bxcommon.constants import LOCALHOST
from bxcommon.messages.bloxroute.ack_message import AckMessage
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxcommon.test_utils import helpers
from bxcommon.test_utils.mocks.mock_bx_messages import hello_message
from bxcommon.test_utils.mocks.mock_node_ssl_service import MockNodeSSLService
from bxcommon.test_utils.mocks.mock_socket_connection import MockSocketConnection

from bxgateway.connections.btc.btc_gateway_node import BtcGatewayNode
from bxgateway.testing.mocks.mock_btc_messages import btc_version_message
from mock import MagicMock


class AbstractBtcGatewayIntegrationTest(AbstractTestCase):
    """
    Abstract test class that setups up two gateways and allow test cases to pass messages back and forth.
    """

    def setUp(self):
        if constants.USE_EXTENSION_MODULES:
            helpers.set_extensions_parallelism()
        self.reinitialize_gateways(self.gateway_1_opts(), self.gateway_2_opts())

    def gateway_1_opts(self):
        return gateway_helpers.get_gateway_opts(9000, peer_gateways=[OutboundPeerModel(LOCALHOST, 7002)],
                                                                  sync_tx_service=False,
                                                                  include_default_btc_args=True)

    def gateway_2_opts(self):
        return gateway_helpers.get_gateway_opts(9001, peer_gateways=[OutboundPeerModel(LOCALHOST, 7002)],
                                                                  sync_tx_service=False,
                                                                  include_default_btc_args=True)

    def reinitialize_gateways(self, opts1, opts2):
        node_ssl_service = MockNodeSSLService(BtcGatewayNode.NODE_TYPE, MagicMock())
        self.node1 = BtcGatewayNode(opts1, node_ssl_service)
        self.node1.opts.has_fully_updated_tx_service = True
        self.node1.requester.send_threaded_request = MagicMock()
        self.node2 = BtcGatewayNode(opts2, node_ssl_service)
        self.node2.opts.has_fully_updated_tx_service = True
        self.node2.requester.send_threaded_request = MagicMock()

        self.node1.peer_gateways = {OutboundPeerModel(LOCALHOST, 7002)}
        self.node1.peer_relays = {OutboundPeerModel(LOCALHOST, 7001)}
        self.node2.peer_gateways = {OutboundPeerModel(LOCALHOST, 7002)}
        self.node2.peer_relays = {OutboundPeerModel(LOCALHOST, 7001)}

        self.blockchain_fileno = 1
        self.relay_fileno = 2
        self.gateway_fileno = 3

        self.blockchain_connection = MockSocketConnection(
            self.blockchain_fileno, self.node1, ip_address=LOCALHOST, port=7000
        )
        self.relay_connection = MockSocketConnection(self.relay_fileno, self.node1, ip_address=LOCALHOST, port=7001)
        self.gateway_connection = MockSocketConnection(self.gateway_fileno, self.node1, ip_address=LOCALHOST, port=7002)

        # add node1 connections
        self.node1.on_connection_added(self.blockchain_connection)
        self.node1.on_connection_added(self.relay_connection)
        self.node1.on_connection_added(self.gateway_connection)

        # add node 2 connections
        self.node2.on_connection_added(self.blockchain_connection)
        self.node2.on_connection_added(self.relay_connection)
        self.node2.on_connection_added(self.gateway_connection)

        # initialize node1 connections
        helpers.receive_node_message(self.node1, self.blockchain_fileno, btc_version_message().rawbytes())
        helpers.receive_node_message(self.node1, self.relay_fileno, hello_message().rawbytes())
        helpers.receive_node_message(self.node1, self.relay_fileno, AckMessage().rawbytes())
        helpers.receive_node_message(self.node1, self.gateway_fileno, AckMessage().rawbytes())

        # initialize node2 connections
        helpers.receive_node_message(self.node2, self.blockchain_fileno, btc_version_message().rawbytes())
        helpers.receive_node_message(self.node2, self.relay_fileno, hello_message().rawbytes())
        helpers.receive_node_message(self.node2, self.relay_fileno, AckMessage().rawbytes())
        helpers.receive_node_message(self.node2, self.gateway_fileno, AckMessage().rawbytes())

        self.clear_all_buffers()

    def clear_all_buffers(self):
        helpers.clear_node_buffer(self.node1, self.blockchain_fileno)
        helpers.clear_node_buffer(self.node1, self.relay_fileno)
        helpers.clear_node_buffer(self.node1, self.gateway_fileno)
        helpers.clear_node_buffer(self.node2, self.blockchain_fileno)
        helpers.clear_node_buffer(self.node2, self.relay_fileno)
        helpers.clear_node_buffer(self.node2, self.gateway_fileno)



import time
from abc import ABCMeta, abstractmethod
from typing import List, Optional

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.models.tx_validation_status import TxValidationStatus
from bxcommon.utils import performance_utils
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.utils.stats.transaction_stat_event_type import TransactionStatEventType
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxgateway import gateway_constants
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.services.gateway_transaction_service import ProcessTransactionMessageFromNodeResult
from bxgateway.utils.stats.gateway_bdn_performance_stats_service import gateway_bdn_performance_stats_service
from bxgateway.utils.stats.gateway_transaction_stats_service import gateway_transaction_stats_service
from bxutils import logging, log_messages
from bxutils.logging.log_record_type import LogRecordType

logger = logging.get_logger(__name__)
msg_handling_logger = logging.get_logger(LogRecordType.MessageHandlingTroubleshooting, __name__)


class AbstractBlockchainConnectionProtocol:
    __metaclass__ = ABCMeta
    connection: AbstractGatewayBlockchainConnection

    def __init__(
        self,
        connection: AbstractGatewayBlockchainConnection,
        block_cleanup_poll_interval_s: int = gateway_constants.BLOCK_CLEANUP_NODE_BLOCK_LIST_POLL_INTERVAL_S,
    ):
        self.block_cleanup_poll_interval_s = block_cleanup_poll_interval_s
        self.connection = connection
        self.node = self.connection.node
        self.tx_service = self.node.get_tx_service()

    def msg_tx(self, msg):
        """
        Handle a TX message by broadcasting to the entire network
        """
        start_time = time.time()
        txn_count = 0
        broadcast_txs_count = 0

        process_tx_msg_result = self.tx_service.process_transactions_message_from_node(
            msg,
            self.node.get_network_min_transaction_fee(),
            self.node.opts.transaction_validation
        )

        self.msg_tx_after_tx_service_process_complete(process_tx_msg_result)

        if not self.node.opts.has_fully_updated_tx_service:
            logger.debug(
                "Gateway received {} transaction messages while syncing, skipping..",
                len(process_tx_msg_result)
            )
            return

        broadcast_start_time = time.time()

        for tx_result in process_tx_msg_result:
            txn_count += 1
            if TxValidationStatus.INVALID_FORMAT in tx_result.tx_validation_status:
                gateway_transaction_stats_service.log_tx_validation_failed_structure()
                logger.warning(
                    log_messages.NODE_RECEIVED_TX_WITH_INVALID_FORMAT,
                    self.node.NODE_TYPE,
                    tx_result.transaction_hash,
                    stats_format.connection(self.connection)
                )

                tx_stats.add_tx_by_hash_event(
                    tx_result.transaction_hash,
                    TransactionStatEventType.TX_VALIDATION_FAILED_STRUCTURE,
                    self.connection.network_num,
                    peers=[self.connection]
                )
                continue

            if TxValidationStatus.INVALID_SIGNATURE in tx_result.tx_validation_status:
                gateway_transaction_stats_service.log_tx_validation_failed_signature()
                logger.warning(
                    log_messages.NODE_RECEIVED_TX_WITH_INVALID_SIG,
                    self.node.NODE_TYPE,
                    tx_result.transaction_hash,
                    stats_format.connection(self.connection)
                )

                tx_stats.add_tx_by_hash_event(
                    tx_result.transaction_hash,
                    TransactionStatEventType.TX_VALIDATION_FAILED_SIGNATURE,
                    self.connection.network_num,
                    peers=[self.connection]
                )
                continue

            if TxValidationStatus.LOW_FEE in tx_result.tx_validation_status:
                gateway_transaction_stats_service.log_tx_validation_failed_gas_price()
                # log low fee transaction here for bdn_performance
                gateway_bdn_performance_stats_service.log_tx_from_blockchain_node(self.connection.endpoint, True)

                logger.trace(
                    "transaction {} has gas price lower then the setting {}",
                    tx_result.transaction_hash,
                    self.node.get_network_min_transaction_fee()
                )

                tx_stats.add_tx_by_hash_event(
                    tx_result.transaction_hash,
                    TransactionStatEventType.TX_VALIDATION_FAILED_GAS_PRICE,
                    self.connection.network_num,
                    peers=[self.connection]
                )
                continue

            if tx_result.seen:
                tx_stats.add_tx_by_hash_event(
                    tx_result.transaction_hash,
                    TransactionStatEventType.TX_RECEIVED_FROM_BLOCKCHAIN_NODE_IGNORE_SEEN,
                    self.connection.network_num,
                    peers=[self.connection]
                )
                gateway_transaction_stats_service.log_duplicate_transaction_from_blockchain()
                gateway_bdn_performance_stats_service.log_duplicate_tx_from_node(self.connection.endpoint)
                continue

            broadcast_txs_count += 1

            tx_stats.add_tx_by_hash_event(
                tx_result.transaction_hash,
                TransactionStatEventType.TX_RECEIVED_FROM_BLOCKCHAIN_NODE,
                self.connection.network_num,
                peers=[self.connection]
            )
            gateway_transaction_stats_service.log_transaction_from_blockchain(tx_result.transaction_hash)

            # log transactions that passed validation, according to fee
            gateway_bdn_performance_stats_service.log_tx_from_blockchain_node(
                self.connection.endpoint,
                not self.node.is_gas_price_above_min_network_fee(tx_result.transaction_contents)
            )

            # All connections outside of this one is a bloXroute server
            broadcast_peers = self.node.broadcast(
                tx_result.bdn_transaction_message,
                self.connection,
                connection_types=(ConnectionType.RELAY_TRANSACTION,)
            )
            self.node.broadcast(
                msg,
                self.connection,
                connection_types=(ConnectionType.BLOCKCHAIN_NODE,)
            )
            gateway_bdn_performance_stats_service.log_tx_sent_to_nodes(broadcasting_endpoint=self.connection.endpoint)

            if self.node.opts.ws:
                self.publish_transaction(
                    tx_result.transaction_hash, memoryview(tx_result.transaction_contents)
                )

            if broadcast_peers:
                tx_stats.add_tx_by_hash_event(
                    tx_result.transaction_hash,
                    TransactionStatEventType.TX_SENT_FROM_GATEWAY_TO_PEERS,
                    self.connection.network_num,
                    peers=broadcast_peers
                )
            else:
                logger.trace(
                    "Tx Message: {} from BlockchainNode was dropped, no upstream relay connection available",
                    tx_result.transaction_hash
                )

        set_content_start_time = time.time()
        end_time = time.time()

        total_duration_ms = (end_time - start_time) * 1000
        duration_before_broadcast_ms = (broadcast_start_time - start_time) * 1000
        duration_broadcast_ms = (set_content_start_time - broadcast_start_time) * 1000
        duration_set_content_ms = (end_time - set_content_start_time) * 1000

        gateway_transaction_stats_service.log_processed_node_transaction(
            total_duration_ms,
            duration_before_broadcast_ms,
            duration_broadcast_ms,
            duration_set_content_ms,
            broadcast_txs_count
        )

        performance_utils.log_operation_duration(
            msg_handling_logger,
            "Process single transaction from Blockchain",
            start_time,
            gateway_constants.BLOCKCHAIN_TX_PROCESSING_TIME_WARNING_THRESHOLD_S,
            connection=self,
            message=msg,
            total_duration_ms=total_duration_ms,
            duration_before_broadcast_ms=duration_before_broadcast_ms,
            duration_broadcast_ms=duration_broadcast_ms,
            duration_set_content_ms=duration_set_content_ms
        )

    def msg_tx_after_tx_service_process_complete(self, process_result: List[ProcessTransactionMessageFromNodeResult]):
        pass

    @abstractmethod
    def msg_block(self, msg: AbstractBlockMessage) -> None:
        """
        Handle a block message. Sends to node for encryption, then broadcasts.
        """
        pass

    def process_msg_block(self, msg: AbstractBlockMessage, block_number: Optional[int] = None) -> None:
        block_hash = msg.block_hash()
        # if gateway is still syncing, skip this process
        if not self.node.should_process_block_hash(block_hash):
            return

        self.node.block_cleanup_service.on_new_block_received(block_hash, msg.prev_block_hash())
        block_stats.add_block_event_by_block_hash(
            block_hash, 
            BlockStatEventType.BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE,
            network_num=self.connection.network_num,
            more_info="Protocol: {}, Network: {}".format(
                self.node.opts.blockchain_protocol,
                self.node.opts.blockchain_network,
                msg.extra_stats_data()
            )
        )
        gateway_bdn_performance_stats_service.log_block_message_from_blockchain_node(self.connection.endpoint, True)
        if block_hash in self.node.blocks_seen.contents:
            self.node.on_block_seen_by_blockchain_node(block_hash, self.connection, msg, block_number=block_number)
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE_IGNORE_SEEN,
                network_num=self.connection.network_num
            )
            self.connection.log_info(
                "Discarding duplicate block {} from local blockchain node.",
                block_hash
            )
            self.node.log_blocks_network_content(self.node.network_num, msg)
            return

        if not self.is_valid_block_timestamp(msg):
            return

        gateway_bdn_performance_stats_service.log_block_from_blockchain_node(self.connection.endpoint)

        canceled_recovery = self.node.on_block_seen_by_blockchain_node(block_hash, self.connection, msg)
        if canceled_recovery:
            return

        self.node.track_block_from_node_handling_started(block_hash)
        self.node.on_block_seen_by_blockchain_node(block_hash, self.connection, msg, block_number=block_number)
        self.node.block_processing_service.queue_block_for_processing(msg, self.connection)
        self.node.block_queuing_service_manager.push(block_hash, msg, node_received_from=self.connection)
        return

    def msg_proxy_request(self, msg, requesting_connection: AbstractGatewayBlockchainConnection):
        """
        Handle a chainstate request message.
        """
        self.node.msg_proxy_requester_queue.append(requesting_connection)
        self.node.send_msg_to_remote_node(msg)

    def msg_proxy_response(self, msg):
        """
        Handle a chainstate response message.
        """
        if self.node.msg_proxy_requester_queue:
            requester_connection = self.node.msg_proxy_requester_queue.popleft()
            requester_connection.enqueue_msg(msg)
        else:
            logger.error(log_messages.MSG_PROXY_REQUESTER_QUEUE_EMPTY_ON_RESPONSE)

    def is_valid_block_timestamp(self, msg: AbstractBlockMessage) -> bool:
        max_time_offset = (
            self.node.opts.blockchain_block_interval *
            self.node.opts.blockchain_ignore_block_interval_count
        )
        if time.time() - msg.timestamp() >= max_time_offset:
            self.connection.log_trace("Received block {} more than {} seconds after it was created ({}). Ignoring.",
                                      msg.block_hash(), max_time_offset, msg.timestamp())
            return False

        return True

    def _request_blocks_confirmation(self):
        if not self.connection.is_alive():
            return None
        last_confirmed_block = self.node.block_cleanup_service.last_confirmed_block
        tracked_blocks = self.tx_service.get_oldest_tracked_block(self.node.network.block_confirmations_count)

        if last_confirmed_block is not None and len(tracked_blocks) <= self.node.network.block_confirmations_count + \
            gateway_constants.BLOCK_CLEANUP_REQUEST_EXPECTED_ADDITIONAL_TRACKED_BLOCKS:
            hashes = [last_confirmed_block]
            hashes.extend(tracked_blocks)
        else:
            hashes = tracked_blocks
        if hashes:
            msg = self._build_get_blocks_message_for_block_confirmation(hashes)
            self.connection.enqueue_msg(msg)
            self.connection.log_debug("Sending block confirmation request. Last confirmed block: {}, hashes: {}",
                                      last_confirmed_block, hashes[:gateway_constants.LOGGING_LIMIT_ITEM_COUNT])
        return self.block_cleanup_poll_interval_s

    @abstractmethod
    def _build_get_blocks_message_for_block_confirmation(self, hashes: List[Sha256Hash]) -> AbstractMessage:
        pass

    def publish_transaction(
        self, tx_hash: Sha256Hash, tx_contents: memoryview
    ) -> None:
        pass



from typing import TYPE_CHECKING, Callable, Optional

from bxcommon.feed.feed import FeedKey
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.requests.subscribe_rpc_request import SubscribeRpcRequest
from bxcommon.feed.feed_manager import FeedManager
from bxcommon.feed.subscriber import Subscriber
from bxcommon.rpc.rpc_errors import RpcInvalidParams

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class GatewaySubscribeRpcRequest(SubscribeRpcRequest):
    def __init__(
        self,
        request: BxJsonRpcRequest,
        node: "AbstractGatewayNode",
        feed_manager: FeedManager,
        subscribe_handler: Callable[[Subscriber, FeedKey, Optional[str]], None],
        feed_network: int = 0
    ) -> None:
        super().__init__(request, node, feed_manager, subscribe_handler, feed_network, node.account_model)

    def validate_params_get_options(self):
        self.feed_network = self.node.network_num
        super().validate_params_get_options()
        if (
            self.feed_name in {rpc_constants.ETH_TRANSACTION_RECEIPTS_FEED_NAME, rpc_constants.ETH_ON_BLOCK_FEED_NAME}
            and (not self.node.opts.ws or not self.node.opts.eth_ws_uri or not self.node.get_ws_server_status())
        ):
            raise RpcInvalidParams(
                self.request_id,
                f"In order to use the {self.feed_name} feed, "
                f"your gateway must be connected to your Ethereum node's websocket server "
                f"and websocket RPC must be enabled on your gateway. "
                f"To do so, start your gateway with the following parameters: "
                f"--ws True "
                f"--ws-host <IP address of client application> "
                f"--ws-port 28333 "
                f"--eth-ws-uri ws://[ip_address]:[port]."
            )



from typing import TYPE_CHECKING

from bxcommon.rpc.https.http_rpc_handler import HttpRpcHandler
from bxcommon.rpc.requests.transaction_status_rpc_request import TransactionStatusRpcRequest
from bxcommon.rpc.rpc_request_type import RpcRequestType
from bxgateway.rpc.requests.add_blockchain_peer_rpc_request import AddBlockchainPeerRpcRequest
from bxgateway.rpc.requests.bdn_performance_rpc_request import BdnPerformanceRpcRequest
from bxgateway.rpc.requests.gateway_blxr_transaction_rpc_request import GatewayBlxrTransactionRpcRequest
from bxgateway.rpc.requests.gateway_memory_usage_report_rpc_request import GatewayMemoryUsageRpcRequest
from bxgateway.rpc.requests.gateway_status_rpc_request import GatewayStatusRpcRequest
from bxgateway.rpc.requests.gateway_stop_rpc_request import GatewayStopRpcRequest
from bxgateway.rpc.requests.gateway_memory_rpc_request import GatewayMemoryRpcRequest
from bxgateway.rpc.requests.gateway_peers_rpc_request import GatewayPeersRpcRequest
from bxgateway.rpc.requests.gateway_transaction_service_rpc_request import GatewayTransactionServiceRpcRequest
from bxgateway.rpc.requests.quota_usage_rpc_request import QuotaUsageRpcRequest
from bxgateway.rpc.requests.gateway_blxr_call_rpc_request import GatewayBlxrCallRpcRequest
from bxgateway.rpc.requests.remove_blockchain_peer_rpc_request import RemoveBlockchainPeerRpcRequest

from bxutils import logging


if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(__name__)


class GatewayHttpRpcHandler(HttpRpcHandler["AbstractGatewayNode"]):

    def __init__(self, node: "AbstractGatewayNode") -> None:
        super().__init__(node)
        self.request_handlers = {
            RpcRequestType.BLXR_TX: GatewayBlxrTransactionRpcRequest,
            RpcRequestType.BLXR_ETH_CALL: GatewayBlxrCallRpcRequest,
            RpcRequestType.GATEWAY_STATUS: GatewayStatusRpcRequest,
            RpcRequestType.STOP: GatewayStopRpcRequest,
            RpcRequestType.MEMORY: GatewayMemoryRpcRequest,
            RpcRequestType.PEERS: GatewayPeersRpcRequest,
            RpcRequestType.BDN_PERFORMANCE: BdnPerformanceRpcRequest,
            RpcRequestType.QUOTA_USAGE: QuotaUsageRpcRequest,
            RpcRequestType.MEMORY_USAGE: GatewayMemoryUsageRpcRequest,
            RpcRequestType.TX_STATUS: TransactionStatusRpcRequest,
            RpcRequestType.TX_SERVICE: GatewayTransactionServiceRpcRequest,
            RpcRequestType.ADD_BLOCKCHAIN_PEER: AddBlockchainPeerRpcRequest,
            RpcRequestType.REMOVE_BLOCKCHAIN_PEER: RemoveBlockchainPeerRpcRequest
        }



import binascii
from math import ceil
from typing import List, Optional

from bxcommon.exceptions import ParseError

# -------
#   Utility functions to work with RLP (Recursive Length Prefix) encoding.
#   https://github.com/ethereum/wiki/wiki/RLP
# -------


def encode_int(value):
    """
    Encode integer value into RLP format

    :param value: int value
    :return: RLP encoded bytes
    """

    if value == 0:
        s = b""
    else:
        s = int_to_big_endian(value)

    if len(s) == 1 and safe_ord(s[0]) < 128:
        return s

    return get_length_prefix_str(len(s)) + s


def consume_length_prefix(rlp, start):
    """
    Ethereum RLP Utils:
    Read a length prefix from an RLP string.

    :param rlp: the rlp string to read from
    :param start: the position at which to start reading
    :returns: a tuple ``(type, length, end)``, where ``type`` is either ``str``
              or ``list`` depending on the type of the following payload,
              ``length`` is the length of the payload in bytes, and ``end`` is
              the position of the first payload byte in the rlp string
    """
    if not isinstance(rlp, memoryview):
        raise TypeError("Only memoryview is allowed for RLP content for best performance. Type provided was: {}"
                        .format(type(rlp)))

    if start is None:
        raise ValueError("Argument start is required")

    first_byte = safe_ord(rlp[start])
    if first_byte < 128:  # single byte
        return (str, 1, start)
    elif first_byte < 128 + 56:  # short string
        if first_byte - 128 == 1 and safe_ord(rlp[start + 1]) < 128:
            raise ParseError("Encoded as short string although single byte was possible")
        return (str, first_byte - 128, start + 1)
    elif first_byte < 192:  # long string
        long_length = first_byte - 128 - 56 + 1
        if rlp[start + 1:start + 2] == b"\x00":
            raise ParseError("Length starts with zero bytes")
        length = big_endian_to_int(rlp[start + 1:start + 1 + long_length])
        if length < 56:
            raise ParseError("Long string prefix used for short string")
        return (str, length, start + 1 + long_length)
    elif first_byte < 192 + 56:  # short list
        return (list, first_byte - 192, start + 1)
    else:  # long list
        long_length = first_byte - 192 - 56 + 1
        if rlp[start + 1:start + 2] == b"\x00":
            raise ParseError("Length starts with zero bytes")
        length = big_endian_to_int(rlp[start + 1:start + 1 + long_length])
        if length < 56:
            raise ParseError("Long list prefix used for short list")
        return (list, length, start + 1 + long_length)


def decode_int(rlp, start):
    """
    Decodes int value from RLP format

    :param rlp: RLP bytes
    :param start: start offset
    :return: tuple (value, length)
    """

    _, value_len, value_start = consume_length_prefix(rlp, start)

    value_bytes = rlp[value_start:value_start + value_len]

    if len(value_bytes) == 0:
        value = 0
    else:
        value = big_endian_to_int(value_bytes)

    return (value, (value_start - start) + value_len)


def get_length_prefix_str(length):
    """
    Calculates length prefix for byte or string

    :param length: length of bytes or string
    :return: length prefix bytes
    """

    if length is None:
        raise ValueError("Argument length is required")

    return get_length_prefix(length, 128)


def get_length_prefix_list(length):
    """
    Calculates length prefix for list

    :param length: length of list
    :return: length prefix bytes
    """

    if length is None:
        raise ValueError("Argument length is required")

    return get_length_prefix(length, 192)


def get_length_prefix(length, offset):
    """Construct the prefix to lists or strings denoting their length.

    :param length: the length of the item in bytes
    :param offset: ``0x80`` when encoding raw bytes, ``0xc0`` when encoding a list
    """

    if length is None:
        raise ValueError("Argument length is required")

    if offset is None:
        raise ValueError("Argument offset is required")

    if length < 56:
        return ascii_chr(offset + length)
    elif length < 256 ** 8:
        length_string = int_to_big_endian(length)
        return ascii_chr(offset + 56 - 1 + len(length_string)) + length_string
    else:
        raise ValueError("Length greater than 256**8")


def int_to_big_endian(value):
    """
    Converts int to big endian

    :param value: int value
    :return: big ending value
    """

    byte_length = max(ceil(value.bit_length() / 8), 1)
    return (value).to_bytes(byte_length, byteorder="big")


def ascii_chr(value):
    """
    Converts byte to ASCII char

    :param value: ASCII code of character
    :return: char
    """

    return bytes([value])


def str_to_bytes(value):
    """
    Converts string to array of bytes

    :param value: string
    :return: array of bytes
    """

    if isinstance(value, (bytes, bytearray, memoryview)):
        return bytes(value)

    return bytes(value, "utf-8")


def get_list_items_bytes(
    list_bytes: memoryview, remove_items_length_prefix: Optional[bool] = False
) -> List[memoryview]:
    """
    Parses items from the list
    :param list_bytes: RLP serialized list bytes
    :param remove_items_length_prefix: indicates if length prefix needs to be removed from each item in the list
    :return: list of bytes
    """
    result = []
    offset = 0

    while offset < len(list_bytes):
        item_type, item_len, item_start = consume_length_prefix(list_bytes, offset)

        # important for transaction encoding, when RLP representation can
        # be either a rather opaque string or a proper RLP tuple
        if item_type == str:
            remove_items_length_prefix = True

        item_bytes = list_bytes[offset:item_start + item_len]
        if remove_items_length_prefix:
            item_bytes = remove_length_prefix(item_bytes)

        result.append(item_bytes)
        offset = item_start + item_len

    assert offset == len(list_bytes)

    return result


def get_first_list_field_items_bytes(
    object_bytes: memoryview, remove_items_length_prefix: Optional[bool] = False
) -> List[memoryview]:
    """
    Parses the first field of RLP serialized object as list
    :param object_bytes: byte of RLP serialized object
    :param remove_items_length_prefix: indicates if length prefix needs to be removed from each item in the list
    :return: list of bytes
    """
    list_bytes = remove_length_prefix(object_bytes)
    return get_list_items_bytes(list_bytes, remove_items_length_prefix)


def remove_length_prefix(item_bytes: memoryview) -> memoryview:
    """
    Removes RLP length prefix from the item bytes
    :param item_bytes: serialized item bytes
    :return: item bytes without length prefix
    """

    _, _, list_itm_start = consume_length_prefix(item_bytes, 0)
    return item_bytes[list_itm_start:]


def safe_ord(character):
    """
    Ethereum RLP Utils:
    Returns an integer representing the Unicode code point of the character or int if int argument is passed

    :param character: character or integer
    :return: integer representing the Unicode code point of the character or int if int argument is passed
    """

    if isinstance(character, int):
        return character
    else:
        return ord(character)


def big_endian_to_int(value):
    """
    Ethereum RLP Utils:
    Convert big endian to int

    :param value: big ending value
    :return: int value
    """

    return int.from_bytes(value, byteorder="big")


def _pack_left(lnum):
    if lnum == 0:
        return b"\0"
    s = hex(lnum)[2:]
    s = s.rstrip("L")
    if len(s) & 1:
        s = "0" + s
    s = binascii.unhexlify(s)
    return s



import uuid

from bxcommon import constants
from bxutils import logging

logger = logging.get_logger(__name__)


def to_bytes(string_input):
    if not string_input:
        return constants.EMPTY_SOURCE_ID

    try:
        raw = uuid.UUID(string_input).bytes
    except ValueError as _e:
        logger.trace("Invalid node ID: {}", string_input)
        return constants.EMPTY_SOURCE_ID
    return raw


def from_bytes(bytes_input):
    if not bytes_input:
        return None
    if not bytes_input.rstrip(constants.MSG_NULL_BYTE):
        return None
    return str(uuid.UUID(bytes=bytes_input))



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.messages.eth.protocol.eth_protocol_message_factory import EthProtocolMessageFactory
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxgateway.messages.eth.protocol.ping_eth_protocol_message import PingEthProtocolMessage
from bxgateway.utils.eth.rlpx_cipher import RLPxCipher
from bxutils import logging

from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(__name__)


class EthBaseConnection(AbstractGatewayBlockchainConnection["EthGatewayNode"], ABC):
    is_handshake_initiator: bool
    rlpx_cipher: RLPxCipher

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "EthGatewayNode"):
        private_key = node.get_private_key()
        public_key = self.connection_public_key(sock, node)
        self.is_handshake_initiator = public_key is not None
        self.rlpx_cipher = RLPxCipher(
            self.is_handshake_initiator, private_key, public_key
        )

        super().__init__(sock, node)

    @abstractmethod
    def connection_public_key(
        self, sock: AbstractSocketConnectionProtocol, node: "EthGatewayNode"
    ) -> bytes:
        pass

    def connection_message_factory(self) -> AbstractMessageFactory:
        factory = EthProtocolMessageFactory(self.rlpx_cipher)
        if self.is_handshake_initiator:
            factory.set_expected_msg_type(EthProtocolMessageType.AUTH_ACK)
        else:
            factory.set_expected_msg_type(EthProtocolMessageType.AUTH)
        return factory

    def ping_message(self) -> AbstractMessage:
        return PingEthProtocolMessage(None)

    def enqueue_msg(self, msg, prepend=False):
        if not self.is_alive():
            return

        self._log_message(msg.log_level(), "Enqueued message: {}", msg)

        full_message_bytes = bytearray()
        for message_bytes in self.connection_protocol.get_message_bytes(msg):
            full_message_bytes.extend(message_bytes)

        self.enqueue_msg_bytes(full_message_bytes, prepend)



from aiohttp.web import Request


class RequestFormatter:
    _request: Request

    def __init__(self, request: Request) -> None:
        self._request = request

    def __repr__(self) -> str:
        return f"HTTPRequest <{self._request.headers}>"



import asyncio
# TODO: remove try-catch when removing py3.7 support
import functools
import gc
from collections import defaultdict, deque
from datetime import datetime

from bxcommon.models.authenticated_peer_info import AuthenticatedPeerInfo

from bxcommon.utils.stats.memory_statistics_service import memory_statistics
from bxgateway.services.block_queuing_service_manager import BlockQueuingServiceManager
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.connections.connection_state import ConnectionState
from bxcommon.connections.internal_node_connection import InternalNodeConnection
from bxcommon.feed.new_block_feed import NewBlockFeed
from bxgateway.utils.stats.transaction_feed_stats_service import transaction_feed_stats_service

try:
    from asyncio.exceptions import CancelledError
except ImportError:
    from asyncio.futures import CancelledError

import time
from abc import ABCMeta, abstractmethod
from concurrent.futures import Future
from typing import Tuple, Optional, ClassVar, Type, Set, List, Iterable, Union, cast, Dict, Deque
from prometheus_client import Gauge

from bxcommon import constants
from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.connections.abstract_node import AbstractNode
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.models.blockchain_network_model import BlockchainNetworkModel
from bxcommon.models.blockchain_peer_info import BlockchainPeerInfo
from bxcommon.models.node_event_model import NodeEventType
from bxcommon.models.node_type import NodeType
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxcommon.models.bdn_account_model_base import BdnAccountModelBase
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.network_direction import NetworkDirection
from bxcommon.network.peer_info import ConnectionPeerInfo
from bxcommon.services import sdn_http_service
from bxcommon.services.broadcast_service import BroadcastService
from bxcommon.storage.block_encrypted_cache import BlockEncryptedCache
from bxcommon.utils import network_latency, memory_utils, convert, node_cache
from bxcommon.utils.alarm_queue import AlarmId
from bxcommon.utils.expiring_dict import ExpiringDict
from bxcommon.utils.expiring_set import ExpiringSet
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import hooks
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.rpc import rpc_constants
from bxcommon.feed.feed_manager import FeedManager
from bxcommon.feed.feed_source import FeedSource
from bxgateway import gateway_constants
from bxgateway import log_messages
from bxgateway.abstract_message_converter import AbstractMessageConverter
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxgateway.connections.gateway_connection import GatewayConnection
from bxcommon.feed.new_transaction_feed import NewTransactionFeed
from bxgateway.gateway_opts import GatewayOpts
from bxgateway.rpc.https.gateway_http_rpc_server import GatewayHttpRpcServer
from bxgateway.services.abstract_block_cleanup_service import AbstractBlockCleanupService
from bxgateway.services.abstract_block_queuing_service import AbstractBlockQueuingService
from bxgateway.services.block_processing_service import BlockProcessingService
from bxgateway.services.block_recovery_service import BlockRecoveryService, RecoveredTxsSource
from bxgateway.services.gateway_broadcast_service import GatewayBroadcastService
from bxgateway.services.gateway_transaction_service import GatewayTransactionService
from bxgateway.services.neutrality_service import NeutralityService
from bxgateway.utils import configuration_utils
from bxgateway.utils.blockchain_message_queue import BlockchainMessageQueue
from bxgateway.utils.logging.status import status_log
from bxgateway.utils.stats.gateway_bdn_performance_stats_service import gateway_bdn_performance_stats_service
from bxgateway.utils.stats.gateway_transaction_stats_service import gateway_transaction_stats_service
from bxgateway.rpc.ws.ws_server import WsServer
from bxgateway.rpc.ipc.ipc_server import IpcServer
from bxutils import logging
from bxutils.services.node_ssl_service import NodeSSLService
from bxutils.ssl.extensions import extensions_factory
from bxutils.ssl.ssl_certificate_type import SSLCertificateType

logger = logging.get_logger(__name__)


class AbstractGatewayNode(AbstractNode, metaclass=ABCMeta):
    """
    bloXroute gateway node. Middlemans messages between blockchain nodes and the bloXroute
    relay network.
    """
    NODE_TYPE = NodeType.EXTERNAL_GATEWAY
    # pyre-fixme[8]: Attribute has type `Type[AbstractRelayConnection]`; used as `None`.
    RELAY_CONNECTION_CLS: ClassVar[Type[AbstractRelayConnection]] = None

    remote_blockchain_ip: Optional[str] = None
    remote_blockchain_port: Optional[int] = None
    remote_node_conn: Optional[AbstractGatewayBlockchainConnection] = None
    remote_blockchain_protocol_version: Optional[int] = None
    remote_blockchain_connection_established: bool = False
    transaction_streamer_peer: Optional[OutboundPeerModel] = None

    _blockchain_liveliness_alarm: Optional[AlarmId] = None
    _relay_liveliness_alarm: Optional[AlarmId] = None
    _active_blockchain_peers_alarm: Optional[AlarmId] = None

    remote_node_msg_queue: BlockchainMessageQueue
    msg_proxy_requester_queue: Deque[AbstractGatewayBlockchainConnection]

    peer_gateways: Set[OutboundPeerModel]
    # if opts.split_relays is set, then this set contains only block relays
    peer_relays: Set[OutboundPeerModel]
    blockchain_peers: Set[BlockchainPeerInfo]
    num_active_blockchain_peers: int
    blocks_seen: ExpiringSet
    in_progress_blocks: BlockEncryptedCache
    block_recovery_service: BlockRecoveryService
    blockchain_peer_to_block_queuing_service: Dict[AbstractGatewayBlockchainConnection, AbstractBlockQueuingService]
    block_storage: ExpiringDict[Sha256Hash, Optional[AbstractBlockMessage]]
    block_queuing_service_manager: BlockQueuingServiceManager
    block_processing_service: BlockProcessingService
    block_cleanup_service: AbstractBlockCleanupService
    _tx_service: GatewayTransactionService

    _block_from_node_handling_times: ExpiringDict[Sha256Hash, float]
    _block_from_bdn_handling_times: ExpiringDict[Sha256Hash, Tuple[float, str]]

    time_blockchain_peer_conn_destroyed_by_ip: Dict[Tuple[str, int], float] = defaultdict(float)

    feed_manager: FeedManager
    has_feed_subscribers: bool

    tracked_block_cleanup_interval_s: float
    account_model: Optional[BdnAccountModelBase]

    def __init__(
        self,
        opts: GatewayOpts,
        node_ssl_service: NodeSSLService,
        tracked_block_cleanup_interval_s=constants.CANCEL_ALARMS
    ) -> None:
        super(AbstractGatewayNode, self).__init__(
            opts, node_ssl_service
        )
        self.opts: GatewayOpts = opts

        if opts.account_model:
            self.account_model = opts.account_model
            node_cache.update_cache_file(
                opts,
                # pyre-fixme[16]: `Optional` has no attribute `account_id`
                accounts={opts.account_model.account_id: self.account_model},
            )
        else:
            node_cache_info = node_cache.read(opts)
            if (
                node_cache_info
                and node_cache_info.accounts
                and node_cache_info.node_model
                and node_cache_info.node_model.account_id is not None
            ):
                # pyre-fixme[6]: Expected `str` for 1st param but got `Optional[str]`
                self.account_model = node_cache_info.accounts[node_cache_info.node_model.account_id]
            else:
                self.account_model = None

        self.quota_level = 0
        self.num_active_blockchain_peers = 0
        self.blockchain_peers = self.get_blockchain_peers_from_cache_and_command_line()
        self.transaction_streamer_peer = None
        self.peer_gateways = set(opts.peer_gateways)
        self.peer_relays = set(opts.peer_relays)
        self.peer_relays_min_count = 1

        self.remote_node_msg_queue = BlockchainMessageQueue(opts.remote_blockchain_message_ttl)
        self.msg_proxy_requester_queue: Deque[AbstractGatewayBlockchainConnection] = deque(
            maxlen=gateway_constants.MSG_PROXY_REQUESTER_QUEUE_LIMIT
        )

        self.blocks_seen = ExpiringSet(
            self.alarm_queue,
            gateway_constants.GATEWAY_BLOCKS_SEEN_EXPIRATION_TIME_S,
            "gateway_blocks_seen"
        )
        self.in_progress_blocks = BlockEncryptedCache(self.alarm_queue)
        self.block_recovery_service = BlockRecoveryService(self.alarm_queue)
        self.neutrality_service = NeutralityService(self)
        self.block_processing_service = BlockProcessingService(self)
        self.block_cleanup_service = self.build_block_cleanup_service()

        self.block_storage = ExpiringDict(
            self.alarm_queue,
            gateway_constants.MAX_BLOCK_CACHE_TIME_S,
            "block_queuing_service_blocks"
        )
        self.blockchain_peer_to_block_queuing_service = {}
        self.block_queuing_service_manager = BlockQueuingServiceManager(
            self.block_storage,
            self.blockchain_peer_to_block_queuing_service
        )

        self.send_request_for_relay_peers_num_of_calls = 0
        self.check_relay_alarm_id: Optional[AlarmId] = None
        self.transaction_sync_start_alarm_id: Optional[AlarmId] = None

        if not self.opts.peer_relays:
            self._schedule_fetch_relays_from_sdn()
        else:
            self._schedule_fetch_relays_from_sdn(
                gateway_constants.RELAY_CONNECTION_REEVALUATION_INTERVAL_S
            )

        if opts.connect_to_remote_blockchain:
            if opts.remote_blockchain_peer is not None:
                self.remote_blockchain_ip = opts.remote_blockchain_ip
                self.remote_blockchain_port = opts.remote_blockchain_port
                self.enqueue_connection(
                    opts.remote_blockchain_ip, opts.remote_blockchain_port, ConnectionType.REMOTE_BLOCKCHAIN_NODE
                )
            else:
                # offset SDN calls so all the peers aren't queued up at the same time
                self.alarm_queue.register_alarm(
                    constants.SDN_CONTACT_RETRY_SECONDS + 1,
                    self.send_request_for_remote_blockchain_peer
                )

        # offset SDN calls so all the peers aren't queued up at the same time
        self.send_request_for_gateway_peers_num_of_calls = 0
        self.alarm_queue.register_alarm(constants.SDN_CONTACT_RETRY_SECONDS + 2, self._send_request_for_gateway_peers)
        self.network = self.get_blockchain_network()

        if opts.use_extensions:
            from bxgateway.services.extension_gateway_transaction_service import ExtensionGatewayTransactionService
            self._tx_service = ExtensionGatewayTransactionService(self, self.network_num)
        else:
            self._tx_service = GatewayTransactionService(self, self.network_num)

        self.init_transaction_stat_logging()
        self.init_bdn_performance_stats_logging()
        self.init_node_config_update()
        self.init_transaction_feed_stat_logging()

        self._block_from_node_handling_times = ExpiringDict(
            self.alarm_queue,
            gateway_constants.BLOCK_HANDLING_TIME_EXPIRATION_TIME_S,
            f"gateway_block_from_node_handling_times",
        )
        self._block_from_bdn_handling_times = ExpiringDict(
            self.alarm_queue,
            gateway_constants.BLOCK_HANDLING_TIME_EXPIRATION_TIME_S,
            f"gateway_block_from_bdn_handling_times",
        )

        self.schedule_blockchain_liveliness_check(opts.initial_liveliness_check)
        self.schedule_relay_liveliness_check(opts.initial_liveliness_check)
        self.schedule_active_blockchain_peers_liveliness_check(
            gateway_constants.ACTIVE_BLOCKCHAIN_PEERS_LIVELINESS_CHECK_S
        )

        self.opts.has_fully_updated_tx_service = False

        self.block_cleanup_processed_blocks = ExpiringSet(
            self.alarm_queue,
            gateway_constants.BLOCK_CONFIRMATION_EXPIRE_TIME_S,
            "gateway_block_cleanup_processed_blocks"
        )

        self.message_converter: Optional[AbstractMessageConverter] = None
        self.account_id: Optional[str] = extensions_factory.get_account_id(
            node_ssl_service.get_certificate(SSLCertificateType.PRIVATE)
        )
        self.default_tx_flag: TransactionFlag = opts.default_tx_flag
        if self.default_tx_flag & TransactionFlag.PAID_TX and self.account_id is None:
            logger.error(log_messages.INVALID_ACCOUNT_ID)
            self.default_tx_flag = TransactionFlag.NO_FLAGS

        self.has_feed_subscribers = False
        self.feed_manager = FeedManager(self)
        self._rpc_server = self.build_rpc_server()
        self._ws_server = self.build_ws_server()
        self._ipc_server = IpcServer(opts.ipc_file, self.feed_manager, self)
        self.init_authorized_live_feeds()

        self.tracked_block_cleanup_interval_s = tracked_block_cleanup_interval_s
        if self.tracked_block_cleanup_interval_s > 0:
            self.alarm_queue.register_alarm(self.tracked_block_cleanup_interval_s, self._tracked_block_cleanup,
                                            alarm_name="tracked_blocks_cleanup")

        self.quota_gauge = Gauge(
            "quota_level",
            "Quota being used",
            ("quota",)
        )
        self.quota_gauge.labels("tx").set_function(
            functools.partial(lambda: int(self.quota_level))
        )
        self.additional_servers = Gauge(
            "additional_servers",
            "Status of additional servers",
            ("servers",)
        )
        self.additional_servers.labels("ws_server").set_function(
            functools.partial(self._ws_server.status)
        )
        self.additional_servers.labels("ipc_server").set_function(
            functools.partial(self._ipc_server.status)
        )
        self.additional_servers.labels("rpc_server").set_function(
            functools.partial(self._rpc_server.status)
        )

        status_log.initialize(
            self.opts.use_extensions,
            str(self.opts.source_version),
            self.opts.external_ip,
            self.opts.continent,
            self.opts.country,
            self.opts.should_update_source_version,
            self.account_id,
            self.quota_level
        )

        if self.opts.should_restart_on_high_memory:
            self.alarm_queue.register_alarm(
                gateway_constants.CHECK_MEMORY_THRESHOLD_INTERVAL_S,
                self._check_memory_threshold,
                alarm_name="check_memory_threshold"
            )

        self.alarm_queue.register_alarm(
            gateway_constants.ONE_DAY_INTERVAL_S,
            self._get_account_record,
            alarm_name="get_count_model"
        )

    @abstractmethod
    def build_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        pass

    @abstractmethod
    def build_relay_connection(self, socket_connection: AbstractSocketConnectionProtocol) -> AbstractRelayConnection:
        pass

    @abstractmethod
    def build_remote_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        pass

    def build_gateway_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> GatewayConnection:
        return GatewayConnection(socket_connection, self)

    @abstractmethod
    def build_block_queuing_service(
        self,
        connection: AbstractGatewayBlockchainConnection
    ) -> AbstractBlockQueuingService:
        pass

    @abstractmethod
    def build_block_cleanup_service(self) -> AbstractBlockCleanupService:
        pass

    def build_rpc_server(self) -> GatewayHttpRpcServer:
        return GatewayHttpRpcServer(self)

    def build_ws_server(self) -> WsServer:
        return WsServer(self.opts.ws_host, self.opts.ws_port, self.feed_manager, self)

    def get_broadcast_service(self) -> BroadcastService:
        return GatewayBroadcastService(self.connection_pool)

    def init_transaction_stat_logging(self) -> None:
        gateway_transaction_stats_service.set_node(self)
        self.alarm_queue.register_alarm(gateway_transaction_stats_service.interval,
                                        gateway_transaction_stats_service.flush_info)

    def init_bdn_performance_stats_logging(self) -> None:
        gateway_bdn_performance_stats_service.set_node(self)
        self.alarm_queue.register_alarm(
            gateway_bdn_performance_stats_service.interval, self.send_bdn_performance_stats
        )

    def init_transaction_feed_stat_logging(self) -> None:
        transaction_feed_stats_service.set_node(self)
        self.alarm_queue.register_alarm(
            transaction_feed_stats_service.interval,
            transaction_feed_stats_service.flush_info
        )

    def init_authorized_live_feeds(self) -> None:
        new_transaction_streaming_valid = False
        account_model = self.account_model
        if account_model:
            new_transaction_streaming = account_model.new_transaction_streaming
            if new_transaction_streaming:
                new_transaction_streaming_valid = new_transaction_streaming.is_service_valid()
        if self.opts.ws and new_transaction_streaming_valid:
            self.init_live_feeds()
        elif self.opts.ws:
            logger.warning(log_messages.TRANSACTION_FEED_NOT_ALLOWED)

    def init_live_feeds(self) -> None:
        self.feed_manager.register_feed(NewTransactionFeed(network_num=self.network_num))
        self.feed_manager.register_feed(NewBlockFeed(network_num=self.network_num))

    def send_bdn_performance_stats(self) -> int:
        relay_connection = next(iter(self.connection_pool.get_by_connection_types((ConnectionType.RELAY_BLOCK,))), None)
        if not relay_connection:
            gateway_bdn_performance_stats_service.create_interval_data_object()
            return gateway_bdn_performance_stats_service.interval

        gateway_bdn_performance_stats_service.close_interval_data()
        relay_connection.send_bdn_performance_stats(gateway_bdn_performance_stats_service.interval_data)
        gateway_bdn_performance_stats_service.create_interval_data_object()
        return gateway_bdn_performance_stats_service.interval

    def init_node_config_update(self) -> None:
        self.update_node_config()
        self.alarm_queue.register_alarm(constants.ALARM_QUEUE_INIT_EVENT, self.update_node_config)

    def update_node_config(self) -> int:
        configuration_utils.update_node_config(self)
        return self.opts.config_update_interval

    def _record_mem_stats(self, include_data_structure_memory: bool = False):
        super(AbstractGatewayNode, self)._record_mem_stats(include_data_structure_memory)
        self._tx_service.log_tx_service_mem_stats(include_data_structure_memory)
        if include_data_structure_memory:
            block_cleanup_service_size = memory_utils.get_special_size(self.block_cleanup_service).size
            hooks.add_obj_mem_stats(
                self.__class__.__name__,
                0,
                self.block_cleanup_service,
                "block_cleanup_service",
                memory_utils.ObjectSize(
                    size=block_cleanup_service_size,
                    flat_size=0,
                    is_actual_size=True),
                object_type=memory_utils.ObjectType.META,
                size_type=memory_utils.SizeType.SPECIAL
            )
            for block_queuing_service in self.block_queuing_service_manager:
                block_queuing_service.log_memory_stats()

    def get_tx_service(self, network_num=None) -> GatewayTransactionService:
        if network_num is not None and network_num != self.opts.blockchain_network_num:
            raise ValueError("Gateway is running with network number '{}' but tx service for '{}' was requested"
                             .format(self.opts.blockchain_network_num, network_num))

        return self._tx_service

    def on_fully_updated_tx_service(self) -> None:
        super().on_fully_updated_tx_service()
        self.requester.send_threaded_request(sdn_http_service.submit_tx_synced_event,
                                             self.opts.node_id)

    def track_block_from_node_handling_started(self, block_hash: Sha256Hash) -> None:
        """
        Tracks that node started handling a block received from blockchain node
        :param block_hash: block hash
        """
        if block_hash not in self._block_from_node_handling_times.contents:
            self._block_from_node_handling_times.add(block_hash, time.time())

    def track_block_from_node_handling_ended(self, block_hash: Sha256Hash) -> float:
        """
        Tracks that node ended handling of a block received from blockchain node and returns total handling duration
        :param block_hash: block hash
        :return: handling duration in ms
        """

        if block_hash not in self._block_from_node_handling_times.contents:
            return 0

        handling_start_time = self._block_from_node_handling_times.contents[block_hash]
        duration = (time.time() - handling_start_time) * 1000
        self._block_from_node_handling_times.remove_item(block_hash)

        return duration

    def track_block_from_bdn_handling_started(self, block_hash: Sha256Hash, relay_description: str) -> None:
        """
        Tracks that node started handling a block received from BDN
        :param block_hash: block hash
        :param relay_description: description of the relay peer where block was received
        """
        if block_hash not in self._block_from_bdn_handling_times.contents:
            self._block_from_bdn_handling_times.add(block_hash, (time.time(), relay_description))

    def track_block_from_bdn_handling_ended(self, block_hash: Sha256Hash) -> Tuple[float, Optional[str]]:
        """
        Tracks that node ended handling of a block received from BDN and returns total handling duration
        :param block_hash: block hash
        :return: Tuple (handling duration in ms, relay peer description)
        """

        if block_hash not in self._block_from_bdn_handling_times.contents:
            return 0, None

        handling_start_time, relay_description = self._block_from_bdn_handling_times.contents[block_hash]
        duration = (time.time() - handling_start_time) * 1000
        self._block_from_bdn_handling_times.remove_item(block_hash)

        return duration, relay_description

    async def init(self) -> None:
        await super(AbstractGatewayNode, self).init()
        if self.opts.rpc:
            try:
                await asyncio.wait_for(
                    self._rpc_server.start(), rpc_constants.RPC_SERVER_INIT_TIMEOUT_S
                )
            except Exception as e:
                logger.error(log_messages.RPC_INITIALIZATION_FAIL, e, exc_info=True)

        if self.opts.ws:
            try:
                await asyncio.wait_for(
                    self._ws_server.start(), rpc_constants.RPC_SERVER_INIT_TIMEOUT_S
                )
            except Exception as e:
                logger.error(log_messages.WS_INITIALIZATION_FAIL, e, exc_info=True)

        if self.opts.ipc:
            try:
                await asyncio.wait_for(
                    self._ipc_server.start(), rpc_constants.RPC_SERVER_INIT_TIMEOUT_S
                )
            except Exception as e:
                logger.error(log_messages.IPC_INITIALIZATION_FAIL, e, exc_info=True)

    async def close(self) -> None:
        try:
            await asyncio.wait_for(self._rpc_server.stop(), rpc_constants.RPC_SERVER_STOP_TIMEOUT_S)
        except (Exception, CancelledError) as e:
            logger.error(log_messages.RPC_CLOSE_FAIL, e, exc_info=True)
        try:
            await asyncio.wait_for(self._ws_server.stop(), rpc_constants.RPC_SERVER_STOP_TIMEOUT_S)
        except (Exception, CancelledError) as e:
            logger.error(log_messages.WS_CLOSE_FAIL, e, exc_info=True)
        try:
            await asyncio.wait_for(self._ipc_server.stop(), rpc_constants.RPC_SERVER_STOP_TIMEOUT_S)
        except (Exception, CancelledError) as e:
            logger.error(log_messages.IPC_CLOSE_FAIL, e, exc_info=True)

        await super(AbstractGatewayNode, self).close()

    def send_request_for_remote_blockchain_peer(self):
        """
        Requests a bloxroute owned blockchain node from the SDN.
        """
        self.requester.send_threaded_request(
            sdn_http_service.fetch_remote_blockchain_peer,
            self.opts.node_id,
            done_callback=self._process_remote_blockchain_peer_from_sdn
        )

    def get_outbound_peer_info(self) -> List[ConnectionPeerInfo]:
        peers = [
            ConnectionPeerInfo(
                IpEndpoint(peer.ip, peer.port),
                convert.peer_node_to_connection_type(self.NODE_TYPE, peer.node_type)
            )
            for peer in self.outbound_peers
        ]
        for blockchain_peer in self.blockchain_peers:
            peers.append(ConnectionPeerInfo(
                IpEndpoint(blockchain_peer.ip, blockchain_peer.port), ConnectionType.BLOCKCHAIN_NODE
            ))
        if self.remote_blockchain_ip is not None and self.remote_blockchain_port is not None:
            peers.append(ConnectionPeerInfo(
                # pyre-fixme[6]: Expected `str` for 1st param but got `Optional[str]`.
                IpEndpoint(self.remote_blockchain_ip, self.remote_blockchain_port),
                ConnectionType.REMOTE_BLOCKCHAIN_NODE)
            )
        return peers

    def build_connection(self, socket_connection: AbstractSocketConnectionProtocol) -> Optional[AbstractConnection]:
        """
        Builds a connection class object based on the characteristics of the ip, port, and direction of the connection.

        For split relays, force shadowing override of the connection's `CONNECTION_TYPE` attribute. This isn't ideal,
        but prevents the necessity of writing many types of connection classes that differ only by the `CONNECTION_TYPE`
        attribute and lots of branching conditional logic.

        :param socket_connection: socket connection accepting or initializing connection
        :return: connection object, if connection class is found
        """
        ip, port = socket_connection.endpoint
        from_me = socket_connection.direction == NetworkDirection.OUTBOUND
        self.alarm_queue.register_approx_alarm(2 * constants.MIN_SLEEP_TIMEOUT, constants.MIN_SLEEP_TIMEOUT,
                                               status_log.update_alarm_callback, self.connection_pool,
                                               self.opts.use_extensions, self.opts.source_version,
                                               self.opts.external_ip, self.opts.continent, self.opts.country,
                                               self.opts.should_update_source_version, self.blockchain_peers,
                                               self.account_id, self.quota_level)
        if self.is_blockchain_peer(ip, port):
            return self.build_blockchain_connection(socket_connection)
        elif self.remote_blockchain_ip == ip and self.remote_blockchain_port == port:
            return self.build_remote_blockchain_connection(socket_connection)
        # only other gateways attempt to actively connect to gateways
        elif not from_me or any(ip == peer_gateway.ip and port == peer_gateway.port
                                for peer_gateway in self.peer_gateways):
            return self.build_gateway_connection(socket_connection)
        elif any(ip == peer_relay.ip and port == peer_relay.port for peer_relay in self.peer_relays):
            relay_connection = self.build_relay_connection(socket_connection)
            relay_connection.format_connection()
            return relay_connection
        else:
            logger.debug("Attempted connection to peer that's not a blockchain, remote blockchain, gateway, or relay. "
                         "Tried: {}:{}, from_me={}. Ignoring.", ip, port, from_me)
            return None

    def is_blockchain_peer(self, ip, port) -> bool:
        return BlockchainPeerInfo(ip, port) in self.blockchain_peers

    def broadcast_transactions_to_nodes(
        self, msg: AbstractMessage, broadcasting_conn: Optional[AbstractConnection]
    ) -> bool:
        self.broadcast(
            msg,
            broadcasting_conn=broadcasting_conn,
            connection_types=(ConnectionType.BLOCKCHAIN_NODE,)
        )
        return True

    def send_msg_to_remote_node(self, msg: AbstractMessage) -> None:
        """
        Sends a message to remote connected blockchain node.
        """
        remote_node_conn = self.remote_node_conn
        if remote_node_conn is not None:
            remote_node_conn.enqueue_msg(msg)
        else:
            logger.trace("Adding message to remote node's message queue: {}", msg)
            self.remote_node_msg_queue.append(msg)

    def continue_retrying_connection(self, ip: str, port: int, connection_type: ConnectionType) -> bool:
        if connection_type == ConnectionType.BLOCKCHAIN_NODE:
            self.num_active_blockchain_peers = len(
                list(
                    self.connection_pool.get_by_connection_types(
                        (ConnectionType.BLOCKCHAIN_NODE,)
                    )
                )
            )
            return self.num_active_blockchain_peers == 0 \
                or ((ip, port) in self.time_blockchain_peer_conn_destroyed_by_ip and
                    time.time() - self.time_blockchain_peer_conn_destroyed_by_ip[(ip, port)]
                    < gateway_constants.MULTIPLE_BLOCKCHAIN_NODE_MAX_RETRY_S)
        else:
            return (super(AbstractGatewayNode, self).continue_retrying_connection(ip, port, connection_type)
                    or OutboundPeerModel(ip, port) in self.opts.peer_gateways
                    or (connection_type == ConnectionType.REMOTE_BLOCKCHAIN_NODE and
                        self.num_retries_by_ip[(ip, port)] < gateway_constants.REMOTE_BLOCKCHAIN_MAX_CONNECT_RETRIES))

    def on_connection_added(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> Optional[AbstractConnection]:
        conn = super().on_connection_added(socket_connection)
        streaming_peer = self.opts.stream_to_peer_gateway
        if (
            conn is None
            or streaming_peer is None
            or conn.CONNECTION_TYPE not in ConnectionType.GATEWAY
        ):
            return conn

        conn = cast(GatewayConnection, conn)
        conn_endpoint = conn.socket_connection.endpoint
        if (
            conn_endpoint.ip_address == streaming_peer.ip
            and conn_endpoint.port == streaming_peer.port
        ):
            conn.stream_confirmation_messages = True

    def on_blockchain_connection_ready(self, connection: AbstractGatewayBlockchainConnection) -> None:
        new_blockchain_peer = BlockchainPeerInfo(connection.peer_ip, connection.peer_port)
        self.blockchain_peers.add(new_blockchain_peer)
        if connection not in self.block_queuing_service_manager.blockchain_peer_to_block_queuing_service:
            block_queuing_service = self.build_block_queuing_service(connection)
            self.block_queuing_service_manager.add_block_queuing_service(connection, block_queuing_service)
        else:
            queuing_service = self.block_queuing_service_manager.blockchain_peer_to_block_queuing_service[connection]
            queuing_service.update_connection(connection)

        self.cancel_blockchain_liveliness_check()
        self.alarm_queue.register_alarm(
            gateway_constants.CHECK_BLOCKCHAIN_CONNECTION_FIRMLY_ESTABLISHED,
            self.check_blockchain_connection_firmly_established,
            new_blockchain_peer
        )
        self.requester.send_threaded_request(
            sdn_http_service.submit_peer_connection_event,
            NodeEventType.BLOCKCHAIN_NODE_CONN_ESTABLISHED,
            self.opts.node_id,
            connection.peer_ip,
            connection.peer_port
        )
        self.num_active_blockchain_peers = len(
            list(
                self.connection_pool.get_by_connection_types(
                    (ConnectionType.BLOCKCHAIN_NODE,)
                )
            )
        )

    def on_blockchain_connection_destroyed(self, connection: AbstractGatewayBlockchainConnection) -> None:
        if BlockchainPeerInfo(connection.peer_ip, connection.peer_port) in self.blockchain_peers:
            event_type = NodeEventType.BLOCKCHAIN_NODE_CONN_ERR
            if (connection.peer_ip, connection.peer_port) not in self.time_blockchain_peer_conn_destroyed_by_ip:
                self.time_blockchain_peer_conn_destroyed_by_ip[(connection.peer_ip, connection.peer_port)] = time.time()
        else:
            event_type = NodeEventType.BLOCKCHAIN_NODE_CONN_REMOVED
            self.time_blockchain_peer_conn_destroyed_by_ip.pop((connection.peer_ip, connection.peer_port), None)

        self.requester.send_threaded_request(
            sdn_http_service.submit_peer_connection_event,
            event_type,
            self.opts.node_id,
            connection.peer_ip,
            connection.peer_port,
            connection.get_connection_state_details()
        )
        self.num_active_blockchain_peers = len(
            list(
                self.connection_pool.get_by_connection_types(
                    (ConnectionType.BLOCKCHAIN_NODE,)
                )
            )
        )

        if self.num_active_blockchain_peers > 0:
            self.block_queuing_service_manager.remove_block_queuing_service(connection)

    def log_refused_connection(self, peer_info: ConnectionPeerInfo, error: str) -> None:
        if peer_info.connection_type == ConnectionType.BLOCKCHAIN_NODE:
            logger.info("Failed to connect to: {}, {}. Verify that provided ip address ({}) and port ({}) "
                        "are correct. Verify that firewall port is open.", peer_info, error,
                        peer_info.endpoint.ip_address, peer_info.endpoint.port)
            self.requester.send_threaded_request(sdn_http_service.submit_peer_connection_event,
                                                 NodeEventType.BLOCKCHAIN_NODE_CONN_ERR,
                                                 self.opts.node_id,
                                                 peer_info.endpoint.ip_address,
                                                 peer_info.endpoint.port,
                                                 "Connection refused")
        elif peer_info.connection_type == ConnectionType.EXTERNAL_GATEWAY:
            logger.debug("Failed to connect to: {}, {}.", peer_info, error)
        else:
            super(AbstractGatewayNode, self).log_refused_connection(peer_info, error)

    def on_remote_blockchain_connection_ready(self, connection: AbstractGatewayBlockchainConnection) -> None:
        for msg in self.remote_node_msg_queue.pop_items():
            connection.enqueue_msg(msg)
        self.remote_node_conn = connection
        self.requester.send_threaded_request(
            sdn_http_service.submit_peer_connection_event,
            NodeEventType.REMOTE_BLOCKCHAIN_CONN_ESTABLISHED,
            self.opts.node_id,
            connection.peer_ip,
            connection.peer_port
        )

    def on_remote_blockchain_connection_destroyed(self, connection: AbstractGatewayBlockchainConnection) -> None:
        self.requester.send_threaded_request(sdn_http_service.submit_peer_connection_event,
                                             NodeEventType.REMOTE_BLOCKCHAIN_CONN_ERR,
                                             self.opts.node_id,
                                             connection.peer_ip,
                                             connection.peer_port)
        self.remote_node_conn = None
        self.remote_node_msg_queue.pop_items()

    def on_relay_connection_ready(self, conn_type: ConnectionType) -> None:
        # when gateway connects to a relay, check if the gateway doesn't have another syncing process and
        # that the connection is relay tx (to avoid double syncing when gateway connects to both block and tx)
        if (
            self.opts.sync_tx_service
            and self.network_num not in self.last_sync_message_received_by_network
            and ConnectionType.RELAY_TRANSACTION in conn_type
            and len(list(self.connection_pool.get_by_connection_types((ConnectionType.RELAY_TRANSACTION,)))) == 1
        ):
            # set sync to false and updating sdn
            self.opts.has_fully_updated_tx_service = False
            self.requester.send_threaded_request(sdn_http_service.submit_tx_not_synced_event, self.opts.node_id)

            alarm_id = self.transaction_sync_start_alarm_id
            if alarm_id:
                self.alarm_queue.unregister_alarm(alarm_id)
            self.transaction_sync_start_alarm_id = \
                self.alarm_queue.register_alarm(constants.FIRST_TX_SERVICE_SYNC_PROGRESS_S, self.sync_tx_services)

        self.cancel_relay_liveliness_check()

    def on_failed_connection_retry(
        self, ip: str, port: int, connection_type: ConnectionType, connection_state: ConnectionState
    ) -> None:
        self.alarm_queue.register_approx_alarm(2 * constants.MIN_SLEEP_TIMEOUT, constants.MIN_SLEEP_TIMEOUT,
                                               status_log.update_alarm_callback, self.connection_pool,
                                               self.opts.use_extensions, self.opts.source_version,
                                               self.opts.external_ip, self.opts.continent, self.opts.country,
                                               self.opts.should_update_source_version, self.blockchain_peers,
                                               self.account_id, self.quota_level)
        if connection_type in ConnectionType.GATEWAY:
            if ConnectionState.ESTABLISHED in connection_state:
                self.requester.send_threaded_request(
                    sdn_http_service.submit_peer_connection_error_event,
                    self.opts.node_id,
                    ip,
                    port
                )
            self._remove_gateway_peer(ip, port)
        elif connection_type == ConnectionType.REMOTE_BLOCKCHAIN_NODE:
            self.send_request_for_remote_blockchain_peer()
        elif ConnectionType.RELAY_BLOCK in connection_type:
            if ConnectionState.ESTABLISHED in connection_state:
                self.requester.send_threaded_request(
                    sdn_http_service.submit_peer_connection_error_event,
                    self.opts.node_id,
                    ip,
                    port
                )
            self.remove_relay_peer(ip, port)

        # Reset number of retries in case if SDN instructs to connect to the same node again
        self.num_retries_by_ip[(ip, port)] = 0

    def on_updated_remote_blockchain_peer(self, outbound_peer) -> int:
        self.remote_blockchain_ip = outbound_peer.ip
        self.remote_blockchain_port = outbound_peer.port
        self.enqueue_connection(outbound_peer.ip, outbound_peer.port, ConnectionType.REMOTE_BLOCKCHAIN_NODE)
        return constants.CANCEL_ALARMS

    def on_block_seen_by_blockchain_node(
        self,
        block_hash: Sha256Hash,
        connection: Optional[AbstractGatewayBlockchainConnection] = None,
        block_message: Optional[AbstractBlockMessage] = None,
        block_number: Optional[int] = None
    ) -> bool:
        if connection:
            block_queuing_service = self.block_queuing_service_manager.get_block_queuing_service(connection)
            if block_queuing_service is not None:
                block_queuing_service.mark_block_seen_by_blockchain_node(
                    block_hash,
                    block_message,
                    block_number
                )

        recovery_canceled = False
        if block_message is not None:
            self.blocks_seen.add(block_hash)
            recovery_canceled = self.block_recovery_service.cancel_recovery_for_block(block_hash)

        if recovery_canceled:
            assert block_message is not None
            self.block_queuing_service_manager.update_recovered_block(block_hash, block_message, connection)
            block_stats.add_block_event_by_block_hash(
                block_hash,
                BlockStatEventType.BLOCK_RECOVERY_CANCELED,
                network_num=self.network_num,
                more_info=RecoveredTxsSource.BLOCK_RECEIVED_FROM_NODE
            )
            logger.debug(
                "Recovery status for block {}: "
                "Block recovery was cancelled by gateway. Reason - {}.",
                block_hash, RecoveredTxsSource.BLOCK_RECEIVED_FROM_NODE
            )

        self.publish_block(
            block_number, block_hash, block_message, FeedSource.BLOCKCHAIN_SOCKET
        )

        self.log_blocks_network_content(self.network_num, block_message)
        return recovery_canceled

    def on_block_received_from_bdn(
        self, block_hash: Sha256Hash, block_message: AbstractBlockMessage
    ) -> None:
        self.blocks_seen.add(block_hash)
        if self.block_queuing_service_manager.get_block_data(block_hash) is None:
            self.block_queuing_service_manager.store_block_data(block_hash, block_message)
        self.block_recovery_service.cancel_recovery_for_block(block_hash)

    def publish_block(
        self,
        block_number: Optional[int],
        block_hash: Sha256Hash,
        block_message: Optional[AbstractBlockMessage],
        source: FeedSource
    ):
        pass

    def log_blocks_network_content(self, network_num: int, block_msg) -> None:
        pass

    def log_txs_network_content(
        self, network_num: int, transaction_hash: Sha256Hash, transaction_contents: Union[bytearray, memoryview]
    ) -> None:
        pass

    def post_block_cleanup_tasks(
        self,
        block_hash: Sha256Hash,
        short_ids: Iterable[int],
        unknown_tx_hashes: Iterable[Sha256Hash]
    ):
        """post cleanup tasks for blocks, override method to implement"""
        pass

    def schedule_blockchain_liveliness_check(self, time_from_now_s: int) -> None:
        if not self._blockchain_liveliness_alarm and self.opts.require_blockchain_connection:
            self._blockchain_liveliness_alarm = self.alarm_queue.register_alarm(time_from_now_s,
                                                                                self.check_blockchain_liveliness)

    def schedule_active_blockchain_peers_liveliness_check(self, time_from_now_s: int) -> None:
        if not self._active_blockchain_peers_alarm and self.opts.require_blockchain_connection:
            self._active_blockchain_peers_alarm = self.alarm_queue.register_alarm(time_from_now_s,
                                                                                  self.check_active_blockchain_peers)

    def schedule_relay_liveliness_check(self, time_from_now_s: int) -> None:
        if not self._relay_liveliness_alarm:
            self._relay_liveliness_alarm = self.alarm_queue.register_alarm(time_from_now_s,
                                                                           self.check_relay_liveliness)

    def cancel_blockchain_liveliness_check(self) -> None:
        blockchain_liveliness_alarm = self._blockchain_liveliness_alarm
        if blockchain_liveliness_alarm:
            self.alarm_queue.unregister_alarm(blockchain_liveliness_alarm)
            self._blockchain_liveliness_alarm = None

    def cancel_relay_liveliness_check(self) -> None:
        relay_liveliness_alarm = self._relay_liveliness_alarm
        if relay_liveliness_alarm:
            self.alarm_queue.unregister_alarm(relay_liveliness_alarm)
            self._relay_liveliness_alarm = None

    def get_blockchain_peers_from_cache_and_command_line(self) -> Set[BlockchainPeerInfo]:
        blockchain_peers = self.opts.blockchain_peers
        blockchain_peers.add(
            BlockchainPeerInfo(self.opts.blockchain_ip, self.opts.blockchain_port, self.opts.node_public_key)
        )
        cache_file_info = node_cache.read(self.opts)
        if cache_file_info is not None:
            blockchain_peers_from_cache = cache_file_info.blockchain_peers
            if blockchain_peers_from_cache is not None:
                for blockchain_peer in blockchain_peers_from_cache:
                    blockchain_peers.add(blockchain_peer)
        return blockchain_peers

    def has_active_blockchain_peer(self) -> bool:
        for conn in self.connection_pool.get_by_connection_types((ConnectionType.BLOCKCHAIN_NODE,)):
            if conn.is_active():
                return True
        return False

    def get_any_active_blockchain_connection(self) -> Optional[AbstractConnection]:
        for conn in self.connection_pool.get_by_connection_types((ConnectionType.BLOCKCHAIN_NODE,)):
            if conn.is_active():
                return conn
        return None

    def check_blockchain_liveliness(self) -> None:
        """
        Checks that the gateway has functional connections to the blockchain node.

        Closes the gateway if it does not, as the gateway is not providing any useful functionality.
        """
        if not self.has_active_blockchain_peer():
            self.should_force_exit = True
            logger.error(log_messages.NO_ACTIVE_BLOCKCHAIN_CONNECTION)

    def check_blockchain_connection_firmly_established(self, blockchain_peer: BlockchainPeerInfo) -> None:
        if self.connection_pool.has_connection(blockchain_peer.ip, blockchain_peer.port):
            self.time_blockchain_peer_conn_destroyed_by_ip.pop((blockchain_peer.ip, blockchain_peer.port), None)

    def check_active_blockchain_peers(self) -> int:
        """
        Checks active blockchain node peers and updates node cache file
        """
        active_blockchain_peers = []
        for peer in self.blockchain_peers:
            if self.connection_pool.has_connection(peer.ip, peer.port):
                active_blockchain_peers.append(peer)
        node_cache.update_cache_file(self.opts, blockchain_peers=active_blockchain_peers)
        return gateway_constants.ACTIVE_BLOCKCHAIN_PEERS_LIVELINESS_CHECK_S

    def check_relay_liveliness(self) -> None:
        if not list(self.connection_pool.get_by_connection_types((ConnectionType.RELAY_ALL,))):
            self.should_force_exit = True
            logger.error(log_messages.NO_ACTIVE_BDN_CONNECTIONS)

    def should_process_block_hash(self, block_hash: Optional[Sha256Hash] = None) -> bool:
        result = True
        reason = None

        if not self.opts.has_fully_updated_tx_service:
            result = False
            reason = "Tx sync in progress"

        if not self.has_active_blockchain_peer():
            result = False
            reason = "No blockchain connection"

        if not result:
            logger.debug(
                "Gateway skipped processing block {}. Reason: {}.",
                block_hash if block_hash is not None else "block message",
                reason
            )
            if block_hash is not None:
                block_stats.add_block_event_by_block_hash(
                    block_hash,
                    BlockStatEventType.ENC_BLOCK_GATEWAY_IGNORE_NO_BLOCKCHAIN,
                    network_num=self.network_num,
                    more_info=reason
                )

        return result

    def reevaluate_transaction_streamer_connection(self) -> None:
        if self.NODE_TYPE is not NodeType.EXTERNAL_GATEWAY:
            return

        if self.transaction_streamer_peer is None:
            logger.error(log_messages.MISSING_TRANSACTION_STREAMER_PEER_INFO)
            return
        transaction_streamer_peer = self.transaction_streamer_peer
        assert transaction_streamer_peer is not None

        now_has_feed_subscribers = self.feed_manager.any_subscribers()
        has_streamer_conn = self.connection_pool.has_connection(
            transaction_streamer_peer.ip,
            transaction_streamer_peer.port,
            transaction_streamer_peer.node_id
        )
        if not self.has_feed_subscribers and now_has_feed_subscribers and not has_streamer_conn:
            self.peer_gateways.add(transaction_streamer_peer)
            self.enqueue_connection(
                transaction_streamer_peer.ip, transaction_streamer_peer.port, convert.peer_node_to_connection_type(
                    self.NODE_TYPE, transaction_streamer_peer.node_type
                )
            )
            self.outbound_peers.add(transaction_streamer_peer)
        elif self.has_feed_subscribers and not now_has_feed_subscribers and has_streamer_conn:
            rem_conn = self.connection_pool.get_by_ipport(
                transaction_streamer_peer.ip, transaction_streamer_peer.port, transaction_streamer_peer.node_id
            )
            if rem_conn:
                rem_conn.mark_for_close(False)
        self.has_feed_subscribers = now_has_feed_subscribers

    def get_network_min_transaction_fee(self) -> int:
        return self.get_blockchain_network().min_tx_network_fee

    def get_blockchain_network(self) -> BlockchainNetworkModel:
        if self.network_num in self.opts.blockchain_networks:
            return self.opts.blockchain_networks[self.network_num]
        raise EnvironmentError(
            f"Unexpectedly did not find network num {self.network_num} in set of blockchain networks: "
            f"{self.opts.blockchain_networks.values()}"
        )

    def sync_tx_services(self) -> int:
        logger.info("Starting to sync transaction state with BDN.")
        super(AbstractGatewayNode, self).sync_tx_services()
        if self.opts.sync_tx_service:
            retry = True
            if self.has_active_blockchain_peer():
                relay_connection: Optional[AbstractRelayConnection] = next(
                    iter(self.connection_pool.get_by_connection_types((ConnectionType.RELAY_ALL,))), None
                )
                if relay_connection and relay_connection.is_active():
                    if self.transaction_sync_timeout_alarm_id:
                        alarm_id = self.transaction_sync_timeout_alarm_id
                        assert alarm_id is not None
                        self.alarm_queue.unregister_alarm(alarm_id)
                        self.transaction_sync_timeout_alarm_id = None

                    self.transaction_sync_timeout_alarm_id = self.alarm_queue.register_alarm(
                        constants.TX_SERVICE_CHECK_NETWORKS_SYNCED_S, self._transaction_sync_timeout
                    )
                    relay_connection.tx_sync_service.send_tx_service_sync_req(self.network_num)
                    self._clear_transaction_service()
                    retry = False
            if retry:
                logger.info("Relay connection is not ready to sync transaction state with BDN. Scheduling retry.")
                return constants.TX_SERVICE_SYNC_PROGRESS_S
        else:
            self.on_fully_updated_tx_service()
        return constants.CANCEL_ALARMS

    def check_sync_relay_connections(self, conn: AbstractConnection) -> int:
        if (
            self.network_num in self.last_sync_message_received_by_network
            and time.time() - self.last_sync_message_received_by_network[self.network_num]
            > constants.LAST_MSG_FROM_RELAY_THRESHOLD_S
        ):
            logger.warning(
                log_messages.RELAY_CONNECTION_TIMEOUT, constants.LAST_MSG_FROM_RELAY_THRESHOLD_S
            )

            # restart the sync process
            conn: InternalNodeConnection = cast(InternalNodeConnection, conn)
            conn.tx_sync_service.send_tx_service_sync_req(self.network_num)

        return constants.LAST_MSG_FROM_RELAY_THRESHOLD_S

    def sync_and_send_request_for_relay_peers(self, network_num: int) -> int:
        return super().sync_and_send_request_for_relay_peers(self.network_num)

    # pyre-fixme[14]: `process_potential_relays_from_sdn` overrides method defined in `AbstractNode` inconsistently.
    def process_potential_relays_from_sdn(self, get_potential_relays_future: Future):
        """
        part of sync_and_send_request_for_relay_peers logic
        This function is called once there is a result from SDN regarding potential_relays.
        The process of potential_relays should be run in the main thread, hence the register_alarm
        :param get_potential_relays_future: the list of potentital relays from the BDN API
        :return:
        """

        try:
            potential_relay_peers = get_potential_relays_future.result()

            if potential_relay_peers:
                node_cache.update_cache_file(self.opts, potential_relay_peers=potential_relay_peers)
            else:
                # if called too many times to sync_and_send_request_for_relay_peers,
                # reset relay peers to empty list in cache file
                if (
                    self.send_request_for_relay_peers_num_of_calls
                    > gateway_constants.SEND_REQUEST_RELAY_PEERS_MAX_NUM_OF_CALLS
                ):
                    node_cache.update_cache_file(self.opts, potential_relay_peers=[])
                    self.send_request_for_relay_peers_num_of_calls = 0
                self.send_request_for_relay_peers_num_of_calls += 1

                cache_file_info = node_cache.read(self.opts)
                if cache_file_info is not None:
                    potential_relay_peers = cache_file_info.relay_peers

                if not potential_relay_peers:
                    self._schedule_fetch_relays_from_sdn()
                    return

            # check the network latency using the thread pool
            self.requester.send_threaded_request(
                self._find_best_relay_peers,
                potential_relay_peers,
                done_callback=self.register_potential_relay_peers_from_future
            )
        except Exception as e:
            logger.info("Got {} when trying to read potential_relays from sdn", e)

    def register_potential_relay_peers_from_future(self, ping_potential_relays_future: Future) -> None:
        best_relay_peers = ping_potential_relays_future.result()
        self._register_potential_relay_peers(best_relay_peers)

    def update_node_settings_from_blockchain_network(self, blockchain_network: BlockchainNetworkModel) -> None:
        self.opts.enable_block_compression = blockchain_network.enable_block_compression

    def get_gateway_transaction_streamers_count(self) -> int:
        return 0

    def is_gas_price_above_min_network_fee(self, transaction_contents: Union[bytearray, memoryview]) -> bool:
        return True

    def on_new_subscriber_request(self) -> None:
        pass

    def init_memory_stats_logging(self):
        memory_statistics.set_node(self)
        memory_statistics.start_recording(
            functools.partial(
                self.record_mem_stats,
                constants.GC_LOW_MEMORY_THRESHOLD,
                constants.GC_MEDIUM_MEMORY_THRESHOLD,
                constants.GC_HIGH_MEMORY_THRESHOLD
            )
        )

    def get_ws_server_status(self) -> bool:
        return self._ws_server.status()

    def _set_transaction_streamer_peer(self) -> None:
        if self.transaction_streamer_peer is not None or self.NODE_TYPE is not NodeType.EXTERNAL_GATEWAY:
            return
        for peer in self.peer_gateways:
            if peer.is_transaction_streamer():
                self.transaction_streamer_peer = peer
                return

    def _process_gateway_peers_from_sdn(self, get_gateway_peers_future: Future) -> None:
        try:
            gateway_peers = get_gateway_peers_future.result()
            if (
                not gateway_peers
                and not self.peer_gateways
                and self.send_request_for_gateway_peers_num_of_calls
                < gateway_constants.SEND_REQUEST_GATEWAY_PEERS_MAX_NUM_OF_CALLS
            ):
                # Try again later
                logger.warning(log_messages.NO_GATEWAY_PEERS)
                self.send_request_for_gateway_peers_num_of_calls += 1
                if (
                    self.send_request_for_gateway_peers_num_of_calls ==
                    gateway_constants.SEND_REQUEST_GATEWAY_PEERS_MAX_NUM_OF_CALLS
                ):
                    logger.warning(log_messages.ABANDON_GATEWAY_PEER_REQUEST)

                self.alarm_queue.register_alarm(
                    constants.SDN_CONTACT_RETRY_SECONDS, self._send_request_for_gateway_peers
                )
            elif gateway_peers:
                logger.debug("Processing updated peer gateways: {}", gateway_peers)
                self._add_gateway_peers(gateway_peers)
                self._set_transaction_streamer_peer()
                self.on_updated_peers(self._get_all_peers())
                self.send_request_for_gateway_peers_num_of_calls = 0
        except Exception as e:
            logger.info("Got {} when trying to read gateway peers from sdn", e)

    def _send_request_for_gateway_peers(self) -> int:
        """
        Requests gateway peers from SDN. Merges list with provided command line gateways.
        """
        self.requester.send_threaded_request(
            sdn_http_service.fetch_gateway_peers,
            self.opts.node_id,
            self.opts.request_remote_transaction_streaming,
            done_callback=self._process_gateway_peers_from_sdn
        )

        return constants.CANCEL_ALARMS

    def _get_all_peers(self) -> Set[OutboundPeerModel]:
        peers = self.peer_gateways.union(self.peer_relays)
        transaction_streamer_peer = self.transaction_streamer_peer
        if transaction_streamer_peer is not None and not self.has_feed_subscribers:
            peers.discard(transaction_streamer_peer)
        return peers

    def _add_gateway_peers(self, gateways_peers: List[OutboundPeerModel]) -> None:
        for gateway_peer in gateways_peers:
            if (
                gateway_peer.ip != self.opts.external_ip
                or gateway_peer.port != self.opts.external_port
            ):
                self.peer_gateways.add(gateway_peer)

    def _remove_gateway_peer(self, ip, port) -> None:
        gateway_to_remove = None
        for peer_gateway in self.peer_gateways:
            if ip == peer_gateway.ip and port == peer_gateway.port:
                gateway_to_remove = peer_gateway
                break

        if gateway_to_remove is not None:
            self.peer_gateways.remove(gateway_to_remove)
            self.outbound_peers = self._get_all_peers()
            if len(self.peer_gateways) < self.opts.min_peer_gateways:
                self.alarm_queue.register_alarm(
                    constants.SDN_CONTACT_RETRY_SECONDS, self._send_request_for_gateway_peers
                )

    def remove_relay_peer(self, ip: str, port: int) -> None:
        """
        Clean up relay peer on connection failure. (after giving up retry)
        Destroys matching transaction relay if split relays enabled.

        :param ip: relay peer ip
        :param port: relay peer port
        """
        self.peer_relays = {peer_relay for peer_relay in self.peer_relays
                            if not (peer_relay.ip == ip and peer_relay.port == port)}
        self.outbound_peers = self._get_all_peers()

        if (
            len(self.peer_relays) < self.peer_relays_min_count
        ):
            logger.debug(
                "Removed relay peer with ip {} and port {}. Current number of relay "
                "peers is lower than required. Scheduling request of new relays from BDN.",
                ip,
                port
            )
            self._schedule_fetch_relays_from_sdn()

        # Reset confirmed blocks if gateway lost connection to block relay
        if len(self.peer_relays) == 0:
            self.block_processing_service.reset_last_confirmed_block_parameters()

    def remove_relay_transaction_peer(self, ip: str, port: int, remove_block_relay: bool = True) -> None:
        """
        Clean up relay transactions peer on connection failure. (after giving up retry)
        Destroys matching block relay.
        :param ip: transaction relay peer ip
        :param port: transaction relay peer port
        :param remove_block_relay: if to remove the corresponding block relay (to avoid infinite recursing)
        """
        if remove_block_relay and self.connection_pool.has_connection(ip, port - 1):
            block_connection = self.connection_pool.get_by_ipport(ip, port - 1)
            logger.debug("Removing relay block connection matching transaction relay host: {}",
                         block_connection)
            block_connection.mark_for_close(False)
            self.remove_relay_peer(ip, port - 1)

    def _find_active_connection(self, outbound_peers):
        for peer in outbound_peers:
            if self.connection_pool.has_connection(peer.ip, peer.port):
                connection = self.connection_pool.get_by_ipport(peer.ip, peer.port)
                if connection.is_active():
                    return connection
        return None

    def _get_next_retry_timeout(self, ip: str, port: int) -> int:
        base_timeout = super(AbstractGatewayNode, self)._get_next_retry_timeout(ip, port)

        if self.is_blockchain_peer(ip, port):
            return base_timeout + gateway_constants.ADDITIONAL_BLOCKCHAIN_RECONNECT_TIMEOUT_S

        return base_timeout

    def _transaction_sync_timeout(self) -> int:
        logger.warning(log_messages.TX_SYNC_TIMEOUT)
        if self.check_sync_relay_connections_alarm_id:
            alarm_id = self.check_sync_relay_connections_alarm_id
            assert alarm_id is not None
            self.alarm_queue.unregister_alarm(alarm_id)
            self.check_sync_relay_connections_alarm_id = None
        self.transaction_sync_timeout_alarm_id = None
        self.on_network_synced(self.network_num)
        self.on_fully_updated_tx_service()
        return constants.CANCEL_ALARMS

    def _find_best_relay_peers(
        self, potential_relay_peers: List[OutboundPeerModel]
    ) -> List[OutboundPeerModel]:
        logger.info("Received list of potential relays from BDN: {}.",
                    ", ".join([node.ip for node in potential_relay_peers]))

        logger.trace("Potential relay peers: {}", [node.node_id for node in potential_relay_peers])

        best_relay_peers = network_latency.get_best_relays_by_ping_latency_one_per_country(
            potential_relay_peers, gateway_constants.MAX_PEER_RELAYS_COUNT, self.peer_relays
        )

        if not best_relay_peers:
            best_relay_peers = potential_relay_peers

        best_relay_country = best_relay_peers[0].get_country()

        if self.opts.min_peer_relays_count is not None and self.opts.min_peer_relays_count > 0:
            self.peer_relays_min_count = self.opts.min_peer_relays_count
        else:
            self.peer_relays_min_count = max(
                gateway_constants.MIN_PEER_RELAYS_BY_COUNTRY[self.opts.country],
                gateway_constants.MIN_PEER_RELAYS_BY_COUNTRY[best_relay_country]
            )

        best_relay_peers = best_relay_peers[:self.peer_relays_min_count]

        logger.trace("Best relay peers to node {} are: {}", self.opts.node_id, best_relay_peers)

        return best_relay_peers

    def _register_potential_relay_peers(self, best_relay_peers: List[OutboundPeerModel]) -> None:
        best_relay_set = set(best_relay_peers)
        for relay in best_relay_set:
            if relay.node_type == NodeType.RELAY_TRANSACTION or relay.node_type == NodeType.RELAY_BLOCK:
                relay.node_type = NodeType.RELAY

        to_disconnect = self.peer_relays - best_relay_set
        to_connect = best_relay_set - self.peer_relays

        if len(to_connect) > 0:
            logger.debug("Sending relay switch notification to SDN.")
            self.requester.send_threaded_request(
                sdn_http_service.submit_gateway_switching_relays_event,
                self.opts.node_id
            )

        for peer_to_remove in to_disconnect:
            logger.info("Disconnecting from current relay {}", peer_to_remove)
            self.peer_relays.remove(peer_to_remove)

        for peer_to_connect in to_connect:
            logger.info("Connecting to better relay {}", peer_to_connect)
            self.peer_relays.add(peer_to_connect)

        self.on_updated_peers(self._get_all_peers())
        self._schedule_fetch_relays_from_sdn(
            gateway_constants.RELAY_CONNECTION_REEVALUATION_INTERVAL_S
        )

    def _tracked_block_cleanup(self) -> float:
        tx_service = self.get_tx_service()
        block_queuing_service = self.block_queuing_service_manager.get_designated_block_queuing_service()
        if block_queuing_service is not None:
            tracked_blocks_to_clean = []
            tracked_blocks = tx_service.get_oldest_tracked_block(0)
            recent_blocks = list(block_queuing_service.iterate_recent_block_hashes())
            for depth, block_hash in enumerate(recent_blocks):
                if depth > self.network.block_confirmations_count and block_hash in tracked_blocks:
                    self.block_cleanup_service.block_cleanup_request(block_hash)
                    tracked_blocks_to_clean.append(block_hash)
            logger.trace(
                "tracked block cleanup, request cleanup of {} blocks: {}, tracked blocks: {} recent blocks: {}",
                len(tracked_blocks_to_clean), tracked_blocks_to_clean, tracked_blocks, recent_blocks)
        else:
            logger.warning(log_messages.TRACKED_BLOCK_CLEANUP_ERROR, "block queuing service is not available")
        return self.tracked_block_cleanup_interval_s

    def _process_remote_blockchain_peer_from_sdn(self, get_remote_blockchain_peer_future: Future):
        try:
            remote_blockchain_peer = get_remote_blockchain_peer_future.result()
            if remote_blockchain_peer is None:
                logger.debug("Did not receive expected remote blockchain peer from BDN. Retrying.")
                self.alarm_queue.register_alarm(
                    gateway_constants.REMOTE_BLOCKCHAIN_SDN_CONTACT_RETRY_SECONDS,
                    self.send_request_for_remote_blockchain_peer
                )
            else:
                logger.debug("Processing remote blockchain peer: {}", remote_blockchain_peer)
                self.on_updated_remote_blockchain_peer(remote_blockchain_peer)
        except Exception as e:
            logger.info("Got {} when trying to read remote blockchain peer from sdn", e)

    def _schedule_fetch_relays_from_sdn(
        self,
        delay: int = constants.SDN_CONTACT_RETRY_SECONDS
    ) -> None:
        """
        Cancels existing tasks for fetching relays from SDN (usually the reevaluation interval)
        and schedules a new one.
        """
        check_relay_alarm_id = self.check_relay_alarm_id
        if check_relay_alarm_id is not None:
            self.alarm_queue.unregister_alarm(check_relay_alarm_id)

        self.check_relay_alarm_id = self.alarm_queue.register_alarm(
            delay, self.sync_and_send_request_for_relay_peers, self.network_num
        )

    def _clear_transaction_service(self) -> None:
        logger.debug("Clearing all data in transaction service.")
        self._tx_service.clear()

    def _check_memory_threshold(self):
        if self.opts.should_restart_on_high_memory and \
                memory_utils.get_app_memory_usage() > gateway_constants.CHECK_MEMORY_THRESHOLD_LIMIT:
            logger.warning(log_messages.NODE_EXCEEDS_MEMORY)
            self.should_force_exit = True
            self.should_restart_on_high_memory = True
        return gateway_constants.CHECK_MEMORY_THRESHOLD_INTERVAL_S

    def _get_account_record(self) -> int:
        if self.account_model:
            self.requester.send_threaded_request(
                sdn_http_service.fetch_account_model,
                self.account_id,
                done_callback=self._update_account_record
            )
        return constants.CANCEL_ALARMS

    def _update_account_record(self, account_model_future: Future):
        account_model = account_model_future.result()
        if account_model:
            if account_model.is_account_valid():
                self.account_model = account_model
                now = datetime.utcnow()
                expire_date = datetime.fromisoformat(account_model.expire_date)
                trigger_alarm = max(
                    int((expire_date - now).total_seconds()) - gateway_constants.ONE_HOUR_INTERVAL_S,
                    gateway_constants.ONE_HOUR_INTERVAL_S
                )
                node_cache.update_cache_file(
                    self.opts,
                    # pyre-fixme[16]: `Optional` has no attribute `account_id`
                    accounts={self.account_model.account_id: self.account_model},
                )
            else:
                trigger_alarm = gateway_constants.ONE_DAY_INTERVAL_S

            self.alarm_queue.register_alarm(
                trigger_alarm,
                self._get_account_record,
                alarm_name="get_count_model"
            )



class ETHError(Exception):
    def __init__(self, msg):
        super(ETHError, self).__init__(msg)

        self.msg = msg


class WrongMACError(ETHError):
    def __init__(self, msg):
        super(WrongMACError, self).__init__(msg)


class InvalidSignatureError(ETHError):
    def __init__(self, msg):
        super(InvalidSignatureError, self).__init__(msg)


class InvalidKeyError(ETHError):
    def __init__(self, msg):
        super(InvalidKeyError, self).__init__(msg)


class AuthenticationError(ETHError):
    def __init__(self, msg):
        super(AuthenticationError, self).__init__(msg)


class DecryptionError(ETHError):
    def __init__(self, msg):
        super(DecryptionError, self).__init__(msg)


class CipherNotInitializedError(ETHError):
    def __init__(self, msg):
        super(CipherNotInitializedError, self).__init__(msg)
